/* psoc-ctsic.c
 *
 * Control driver for PSOC Capacitive TouchScreen IC
 *
 * Copyright (C) 2006 TomTom BV <http://www.tomtom.com/>
 * Authors: Vincent Dejouy <vincent.dejouy@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/cdev.h>
#include <linux/i2c.h>
#include <linux/workqueue.h>
#include <linux/spinlock.h>
#include <linux/err.h>
#include <linux/psoc-ctsic.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/firmware.h>
#include <linux/kernel.h>
#include <linux/moduleparam.h>
#include <linux/jiffies.h>
#include <plat/fdt.h>

#define VERBOSE_LEVEL    VERBOSE_ERROR

static struct _psoc_i2c {
	struct i2c_client	*client;
	struct work_struct	update_work;
	struct work_struct	sys_info_work;
	struct work_struct	timeout_work;
} psoc_bootloader_i2c;

I2C_DATA_RECEIVED_SYS_INFO_STRUCT i2c_packet_received_sys_info;

static int current_mode = OPERATING_MODE;
static int len_i2c_frame_sys_info;
static struct firmware psoc_firmware;
static int index_firmware=0;
static char current_state = STATE_INITIAL;
static char buff_temp_ascii[MAX_CHAR_PER_LINE];
static char buff_temp_byte[MAX_BYTE_PER_LINE];
static psoc_pdata_t    pdata;
static psoc_drv_data_t *drv_data;
static struct timer_list mytimer;
static int interrupt_triggered=0;

/**
 * psoc_bootloader_i2c_recv - issue a single I2C message in master receive mode
 * @buf: Where to store data read from slave
 * @count: How many bytes to read
 *
 * Returns negative errno, or else the number of bytes read.
 */
static int psoc_bootloader_i2c_recv(char *buf, int count, int verbose)
{
	int ret = -ENODEV;
	int i;

	BUG_ON(!buf);
	ret = i2c_master_recv(psoc_bootloader_i2c.client, buf, count);

	if ( (ret != count) && verbose )
		printk(KERN_ERR PSOC_DEVNAME ": Error reading bootloader registers, return value of i2c read: [%d] instead of [%d]\n", ret, count);
	
	if (verbose == VERBOSE_DEBUG) {
		printk("buff received from bootloader of cypress cap touch screen:\n");
		for (i=0;i<count;i++) {
			printk("0x%x ", buf[i]);
		}
		printk("\n");
		printk("Result of recv i2c cmd: %d\n", ret);
	}

	return ret;
}

/**
 * psoc_bootloader_i2c_send - issue a single I2C message in master receive mode
 * @buf: Message to be written to the slave
 * @count: How many bytes to read
 *
 * Returns negative errno, or else the number of bytes read.
 */
static int psoc_bootloader_i2c_send(char *buf, int count, int verbose)
{
	int ret = -ENODEV;
	int i;

	BUG_ON(!buf);

	ret = i2c_master_send(psoc_bootloader_i2c.client, buf, count);
	if ( (ret != count) && verbose )
		printk(KERN_ERR PSOC_DEVNAME ": Error writing bootloader registers, return value of i2c send: [%d] instead of [%d]\n", ret,count);

	mdelay(10);

	if (verbose == VERBOSE_DEBUG) {
		printk("buff sent to bootloader of cypress cap touch screen:\n");
		for (i=0;i<count;i++) {
			printk("0x%x ", buf[i]);
		}
		printk("\n");
		printk("Result of send i2c cmd: %d\n", ret);
	}

	return ret;
}

static int psoc_bootloader_read_status(int verbose)
{
	int ret;
	char reg_addr = STATUS_REGISTER;
	char buff[3];
	int i;

	ret = psoc_bootloader_i2c_send(&reg_addr, sizeof(char), verbose);
	if ( ret != sizeof(char) ){
		printk(KERN_ERR PSOC_DEVNAME ": Error writing 1 byte to the touch screen\n");
		return -1;
	}

	udelay(200);
	
	ret = psoc_bootloader_i2c_recv(buff, 3*sizeof(char), verbose);
	if ( ret != 3*sizeof(char) ){
		printk(KERN_ERR PSOC_DEVNAME ": Error reading 3 bytes from the touch screen\n");
		return -1;
	}

	if (verbose == VERBOSE_DEBUG){
		printk("Result from getting status:\n");
		for (i=0;i<3;i++)
		{
			printk("0x%x ", buff[i]);
		}

		printk("\nStatus Read = 0x%X\n", buff[2]);
	}

	if ( buff[2] == 0x20 ){
		return 0;
	} else {
		printk(KERN_ERR PSOC_DEVNAME ": Status incorrect, Result from getting status:\n");
		for (i=0;i<3;i++)
			printk("0x%x ", buff[i]);

		printk("\nStatus Read = 0x%X\n", buff[2]);
		return -1;
	}

}

//Returns 0 if erase not finished
//Returns 1 if erase finished
static int psoc_bootloader_flash_erase_finished(int verbose)
{
	int ret;
	char reg_addr = STATUS_REGISTER;
	char buff[3];
	int i;

	ret = psoc_bootloader_i2c_send(&reg_addr, sizeof(char), verbose);
	if ( ret != sizeof(char) ){
		printk(KERN_ERR PSOC_DEVNAME ": Error writing 1 byte to the touch screen\n");
		return -1;
	}

	udelay(200);
	
	ret = psoc_bootloader_i2c_recv(buff, 3*sizeof(char), verbose);
	if ( ret != 3*sizeof(char) ){
		printk(KERN_ERR PSOC_DEVNAME ": Error reading 3 bytes from the touch screen\n");
		return -1;
	}

	if (verbose == VERBOSE_DEBUG){
		printk("Result from getting flash erase status:\n");
		for (i=0;i<3;i++)
		{
			printk("0x%x ", buff[i]);
		}

		printk("\nFlash erase status Read = 0x%X\n", buff[1]);
	
	}

	if ( ( buff[1] & 0x80 ) != 0 ){
		return 0;
	} else {
		return 1;
	}
}


static bool psoc_bootloader_isadigit(char c)
{
	if ( ( c >= '0' ) && ( c <= '9' ) )
		return true;
	else
		return false;
}

static char psoc_bootloader_atoc(char* buff)
{
	char res;

	if ( psoc_bootloader_isadigit(*(buff+1)) == true )
		res = *(buff+1) - '0';
	else
		res = *(buff+1) - 'A' + 10;

	if ( psoc_bootloader_isadigit(*(buff)) == true )
		res += (*(buff) - '0')*16;
	else
		res += (*(buff) - 'A' + 10)*16;

	return res;

}

static int psoc_bootloader_extract_line(char* buffer, char* last_line, int verbose)
{
	int i=0;

	while( (*psoc_firmware.data != '\n') && (*psoc_firmware.data != '\r') && (i < (MAX_CHAR_PER_LINE - 2)) ){
		if (index_firmware == (int) psoc_firmware.size) {
			if (verbose == VERBOSE_DEBUG)
				printk("end of file detected\n");
			*last_line = 1;
			break;
		}
		buffer[i++]=*(psoc_firmware.data++);
		index_firmware++;
	}

	buffer[i]=0;

	if (!i){
		psoc_firmware.data++;
		index_firmware++;
	}

	return i;
}

static void psoc_bootloader_convert_short_to_buff(short number_to_convert, char** output_buffer)
{
	(*output_buffer)[0]=(char)(number_to_convert >> 8);
	(*output_buffer)[1]=(char)(number_to_convert);
}

static int psoc_bootloader_state_generic(char* buff, int length, int verbose)
{

	int i,j;
	char buff_to_send[MAX_LENGTH_I2C_FRAME + SIZE_PREFIX_BOOTLOADER_CMD];
	int length_last_line;
	int length_line;
	int quantity_of_lines;
	short prefix=0;
	char buff_prefix[SIZE_PREFIX_BOOTLOADER_CMD];
	int addr_buff_prefix = (int) buff_prefix;

	length_last_line = length%MAX_LENGTH_I2C_FRAME;
	quantity_of_lines = (length/MAX_LENGTH_I2C_FRAME) + (length_last_line ? 1: 0);

	for (i=0;i<quantity_of_lines;i++)
	{
		psoc_bootloader_convert_short_to_buff(prefix, (char**) &addr_buff_prefix);
		buff_to_send[0]=buff_prefix[0];
		buff_to_send[1]=buff_prefix[1];
		
		if ( i != (quantity_of_lines-1))
			length_line=MAX_LENGTH_I2C_FRAME;
		else
			length_line=length_last_line;

		for (j=0;j<length_line;j++)
			buff_to_send[j+SIZE_PREFIX_BOOTLOADER_CMD] = buff[i*MAX_LENGTH_I2C_FRAME + j];

		psoc_bootloader_i2c_send(buff_to_send, length_line+SIZE_PREFIX_BOOTLOADER_CMD, verbose);
		prefix+=0x10;
	}

	return 0;

}

static int psoc_bootloader_state_write_block(char* buff, int length, int verbose)
{
	int ret;

	if (verbose == VERBOSE_DEBUG){
		printk("Write Block %.2X%.2X\n", buff[10], buff[11]);
	}
	psoc_bootloader_state_generic(buff,length,verbose);
	msleep(100);

	ret = psoc_bootloader_read_status(verbose);
	if ( ret )
		printk(KERN_ERR PSOC_DEVNAME ": Error of status in state write block\n");

	return ret;
}

static int psoc_bootloader_state_enter_bootloader(char* buff, int length, int verbose)
{
	int ret;
	int retries;
	
	if (verbose == VERBOSE_DEBUG){
		printk("Enter Bootloader Mode\n");
		printk("Key: %.2X%.2X%.2X%.2X%.2X%.2X%.2X%.2X\n", buff[2], buff[3], buff[4], buff[5], buff[6], buff[7], buff[8], buff[9]);
	}
	psoc_bootloader_state_generic(buff,length,verbose);
	msleep(300);

	retries=20;
	while ( !psoc_bootloader_flash_erase_finished(VERBOSE_DISABLED) )
	{
		msleep(300);
		retries--;
		if ( !retries )
			break;
	}


	ret = psoc_bootloader_read_status(verbose);
	if ( ret )
		printk(KERN_ERR PSOC_DEVNAME ": Error of status in state enter bootloader\n");

	return ret;
}

static int psoc_bootloader_state_exit_bootloader(char* buff, int length, int verbose)
{
	if (verbose == VERBOSE_DEBUG) {
		printk("Exit Bootloader Mode\n");
		printk("Key: %.2X%.2X%.2X%.2X%.2X%.2X%.2X%.2X\n", buff[2], buff[3], buff[4], buff[5], buff[6], buff[7], buff[8], buff[9]);
	}
	psoc_bootloader_state_generic(buff,length,verbose);

	return 0;
}

static int psoc_bootloader_execute_state(char state, char* buffer, int length, int verbose)
{
	int ret=0;

	switch (state) {
		case STATE_ENTER_BOOTLOADER: ret=psoc_bootloader_state_enter_bootloader(buffer, length, verbose);
					break;
		case STATE_WRITE_BLOCK: ret=psoc_bootloader_state_write_block(buffer, length, verbose);
					break;
		case STATE_EXIT_BOOTLOADER: ret=psoc_bootloader_state_exit_bootloader(buffer, length, verbose);
					break;
		default: ret=ERROR_UNKNOWN_STATE;
	}

	return ret;
}


static int psoc_bootloader_process_line(char* buffer, int length, int verbose)
{
	char cmd;

	if ( buffer[0] != BOOTLOADER_CMD_PREFIX )
		return ERROR_INVALID_FIRMWARE_FORMAT;

	cmd = buffer[1];

	switch (cmd) {
		case ENTER_BOOTLOADER_CMD:
		case WRITE_BLOCK_CMD:
		case EXIT_BOOTLOADER_CMD: break;
		default: return ERROR_UNKNOWN_CMD;
	}

	switch (current_state) {
		case STATE_INITIAL: 		if ( cmd == ENTER_BOOTLOADER_CMD )
							current_state = STATE_ENTER_BOOTLOADER;
						else
							return ERROR_FIRMWARE_DOES_NOT_START_WITH_ENTER_BL_CMD;
						break;
		case STATE_ENTER_BOOTLOADER: 	if ( cmd == ENTER_BOOTLOADER_CMD )
							current_state = STATE_ENTER_BOOTLOADER;
						else if ( cmd == WRITE_BLOCK_CMD )
							current_state = STATE_WRITE_BLOCK;
						else
							return ERROR_INVALID_CMD_SEQUENCE;
						break;
		case STATE_WRITE_BLOCK: 	if ( cmd == WRITE_BLOCK_CMD )
							current_state = STATE_WRITE_BLOCK;
						else if ( cmd == EXIT_BOOTLOADER_CMD )
							current_state = STATE_EXIT_BOOTLOADER;
						else
							return ERROR_INVALID_CMD_SEQUENCE;
						break;
		case STATE_EXIT_BOOTLOADER: 	if ( cmd == EXIT_BOOTLOADER_CMD )
							current_state = STATE_EXIT_BOOTLOADER;
						else
							return ERROR_INVALID_CMD_SEQUENCE;
						break;
		default:			return ERROR_UNKNOWN_STATE;
	}

	if (verbose == VERBOSE_DEBUG)
		printk("\nExecute state %d\n", current_state);
	return psoc_bootloader_execute_state(current_state, buffer, length, verbose);

}

static int psoc_bootloader_parse_lines(int verbose)
{
	int nr_chars_line = 0;
	int nr_bytes_line=0;
	char flag_last_line=0;
	int i,j,k;
	int ret;

	k=0;

	while (!flag_last_line){

		nr_chars_line = psoc_bootloader_extract_line(buff_temp_ascii, &flag_last_line, verbose);

		if (nr_chars_line) {

			nr_bytes_line=nr_chars_line/2;
			j=0;
			for (i=0;i<nr_chars_line;i+=2)
				buff_temp_byte[j++] = psoc_bootloader_atoc( buff_temp_ascii + i );

			ret = psoc_bootloader_process_line(buff_temp_byte, nr_bytes_line, verbose);

			if (ret){
				printk(KERN_ERR PSOC_DEVNAME ": Error processing line %d\n", k);
				return -1;
			}

			k++;

		}
	}

	return 0;

}

static inline unsigned short psoc_bootloader_swap_lsb_msb(char* number_to_convert)
{
	return ( (number_to_convert[1]) | ( number_to_convert[0] << 8) );
}

static int psoc_bootloader_get_system_info(int verbose)
{

	int ret;
	int i;
	char* buff_temp;
	unsigned short temp;

	if (verbose == VERBOSE_DEBUG)
		printk("Enter in get sys info\n");

	buff_temp = (char*) kmalloc(len_i2c_frame_sys_info, GFP_KERNEL);
	if (buff_temp == NULL){
		if (verbose)
			printk(KERN_ERR PSOC_DEVNAME ": Error reading system information, can't allocate a buffer of [%d] bytes\n", len_i2c_frame_sys_info);
		return 1;
	}

	ret = psoc_bootloader_i2c_recv(buff_temp, len_i2c_frame_sys_info, verbose);

	if (ret != len_i2c_frame_sys_info){
		if (verbose)
			printk(KERN_ERR PSOC_DEVNAME ": Error reading system information, return value of i2c read: [%d] instead of [%d]\n", ret, len_i2c_frame_sys_info);
		kfree(buff_temp);
		return 1;
	}

	if (verbose == VERBOSE_DEBUG){
		printk("\n");
		printk(PSOC_DEVNAME ":Get system info, raw data received:\n");
		for (i=0;i<len_i2c_frame_sys_info;i++){
			printk("%.2x ",buff_temp[i]);
		}
	}

	memcpy((void*) &i2c_packet_received_sys_info, (void*) buff_temp, len_i2c_frame_sys_info);

	if (verbose == VERBOSE_DEBUG){
		printk(PSOC_DEVNAME ":i2c packet system info received before resolution reversed:\n");
		printk("HST_MODE: 0x%x\n", i2c_packet_received_sys_info.hst_mode);
		printk("BIST_COMM: 0x%x\n", i2c_packet_received_sys_info.bist_comm);
		printk("BIST_STAT: 0x%x\n", i2c_packet_received_sys_info.bist_stat);

		printk("SILICON_ID0: 0x%x\n", i2c_packet_received_sys_info.siliconID0);
		printk("SILICON_ID1: 0x%x\n", i2c_packet_received_sys_info.siliconID1);
		printk("SILICON_ID2: 0x%x\n", i2c_packet_received_sys_info.siliconID2);
		printk("SILICON_ID3: 0x%x\n", i2c_packet_received_sys_info.siliconID3);
		printk("SILICON_ID4: 0x%x\n", i2c_packet_received_sys_info.siliconID4);
		printk("SILICON_ID5: 0x%x\n", i2c_packet_received_sys_info.siliconID5);
		printk("SILICON_ID6: 0x%x\n", i2c_packet_received_sys_info.siliconID6);
		printk("SILICON_ID7: 0x%x\n", i2c_packet_received_sys_info.siliconID7);

		printk("BOOTLOADER_VERSION: 0x%x\n", i2c_packet_received_sys_info.bootloader_version);
		printk("TRUETOUCH_STANDARD_PRODUCT_VERSION: 0x%x\n", i2c_packet_received_sys_info.truetouch_standard_product_version);
		printk("APPLICATION_ID: 0x%x\n", i2c_packet_received_sys_info.applicationID);
		printk("APPLICATION_VERSION_MAJOR: 0x%x\n", i2c_packet_received_sys_info.application_version_major);
		printk("APPLICATION_VERSION_MINOR: 0x%x\n", i2c_packet_received_sys_info.application_version_minor);

		printk("RESOLUTION_X: 0x%x\n", i2c_packet_received_sys_info.resolution_X);
		printk("RESOLUTION_Y: 0x%x\n", i2c_packet_received_sys_info.resolution_Y);
		printk("NUMBER_OF_GESTURES: 0x%x\n", i2c_packet_received_sys_info.number_gestures);

		printk("\n");
	}

	temp = psoc_bootloader_swap_lsb_msb((char*) &i2c_packet_received_sys_info.bootloader_version);
	i2c_packet_received_sys_info.bootloader_version = temp;
	temp = psoc_bootloader_swap_lsb_msb((char*) &i2c_packet_received_sys_info.truetouch_standard_product_version);
	i2c_packet_received_sys_info.truetouch_standard_product_version = temp;
	temp = psoc_bootloader_swap_lsb_msb((char*) &i2c_packet_received_sys_info.applicationID);
	i2c_packet_received_sys_info.applicationID = temp;

	temp = psoc_bootloader_swap_lsb_msb((char*) &i2c_packet_received_sys_info.resolution_X);
	i2c_packet_received_sys_info.resolution_X = temp;
	temp = psoc_bootloader_swap_lsb_msb((char*) &i2c_packet_received_sys_info.resolution_Y);
	i2c_packet_received_sys_info.resolution_Y = temp;

	if (verbose == VERBOSE_DEBUG){

		printk(PSOC_DEVNAME ":i2c packet system info received before resolution reversed:\n");
		printk("HST_MODE: 0x%x\n", i2c_packet_received_sys_info.hst_mode);
		printk("BIST_COMM: 0x%x\n", i2c_packet_received_sys_info.bist_comm);
		printk("BIST_STAT: 0x%x\n", i2c_packet_received_sys_info.bist_stat);

		printk("SILICON_ID0: 0x%x\n", i2c_packet_received_sys_info.siliconID0);
		printk("SILICON_ID1: 0x%x\n", i2c_packet_received_sys_info.siliconID1);
		printk("SILICON_ID2: 0x%x\n", i2c_packet_received_sys_info.siliconID2);
		printk("SILICON_ID3: 0x%x\n", i2c_packet_received_sys_info.siliconID3);
		printk("SILICON_ID4: 0x%x\n", i2c_packet_received_sys_info.siliconID4);
		printk("SILICON_ID5: 0x%x\n", i2c_packet_received_sys_info.siliconID5);
		printk("SILICON_ID6: 0x%x\n", i2c_packet_received_sys_info.siliconID6);
		printk("SILICON_ID7: 0x%x\n", i2c_packet_received_sys_info.siliconID7);

		printk("BOOTLOADER_VERSION: 0x%x\n", i2c_packet_received_sys_info.bootloader_version);
		printk("TRUETOUCH_STANDARD_PRODUCT_VERSION: 0x%x\n", i2c_packet_received_sys_info.truetouch_standard_product_version);
		printk("APPLICATION_ID: 0x%x\n", i2c_packet_received_sys_info.applicationID);
		printk("APPLICATION_VERSION_MAJOR: 0x%x\n", i2c_packet_received_sys_info.application_version_major);
		printk("APPLICATION_VERSION_MINOR: 0x%x\n", i2c_packet_received_sys_info.application_version_minor);
		
		printk("RESOLUTION_X: 0x%x\n", i2c_packet_received_sys_info.resolution_X);
		printk("RESOLUTION_Y: 0x%x\n", i2c_packet_received_sys_info.resolution_Y);
		printk("NUMBER_OF_GESTURES: 0x%x\n", i2c_packet_received_sys_info.number_gestures);

		printk("\n");

	}

	printk(PSOC_DEVNAME ":Cypress Capacitive Touch Screen System Information:\n");
                
	printk("Silicon ID0                         : 0x%x\n", i2c_packet_received_sys_info.siliconID0);
	printk("Silicon ID1                         : 0x%x\n", i2c_packet_received_sys_info.siliconID1);

	printk("Bootloader Version                  : 0x%x\n", i2c_packet_received_sys_info.bootloader_version);
	printk("TrueTouch Standard Product Version  : 0x%x\n", i2c_packet_received_sys_info.truetouch_standard_product_version);
	printk("Application ID                      : 0x%x\n", i2c_packet_received_sys_info.applicationID);
	printk("Application Version                 : %x.%.2x\n", i2c_packet_received_sys_info.application_version_major, i2c_packet_received_sys_info.application_version_minor);

	printk("Resolution X                        : 0x%x\n", i2c_packet_received_sys_info.resolution_X);
	printk("Resolution Y                        : 0x%x\n", i2c_packet_received_sys_info.resolution_Y);
	printk("Number of Gestures Supported        : %d\n", i2c_packet_received_sys_info.number_gestures);

	kfree(buff_temp);

	if (verbose == VERBOSE_DEBUG)
		printk("Exit from get sys info\n");

	return 0;

}
static int psoc_bootloader_set_offset_register(int verbose, char offset)
{
	int ret;

	ret = psoc_bootloader_i2c_send(&offset, 1, verbose);

	if (verbose == VERBOSE_DEBUG){
		printk(PSOC_DEVNAME ":set offset register to:\n");
		printk("0x%x ",offset);
		printk("\n");
	}

	if (ret != 1 && verbose)
		printk(KERN_ERR PSOC_DEVNAME ": Error setting offset register, return value of i2c send: [%d] instead of 1\n", ret);
          
	return (ret == 1)?0:1;

}

static int psoc_bootloader_read_register(int verbose, char* value)
{
	int ret;

	//A register is 1 byte so we read 1 byte
	ret = psoc_bootloader_i2c_recv(value, 1, verbose);

	if (verbose == VERBOSE_DEBUG){
		printk(PSOC_DEVNAME ":register value read:\n");
		printk("0x%x ",*value);
		printk("\n");
	}

	if (ret != 1 && verbose)
		printk(KERN_ERR PSOC_DEVNAME ": Error reading register value, return value of i2c read: [%d] instead of 1\n", ret);

	return (ret == 1)?0:1;

}

static int psoc_bootloader_write_register(int verbose, char reg, char value)
{
	int ret;
	char buff[2];

	buff[0]=reg;
	buff[1]=value;

	ret = psoc_bootloader_i2c_send(buff, 2, verbose);

	if (verbose == VERBOSE_DEBUG){
		printk(PSOC_DEVNAME ":register offset = 0x%X value set: 0x%X\n", reg, value);
	}

	if (ret != 2 && verbose)
		printk(KERN_ERR PSOC_DEVNAME ": Error setting register value, return value of i2c send: [%d] instead of 2\n", ret);

	return (ret == 2)?0:1;

}
static int psoc_bootloader_switch_to_other_mode(int verbose, char device_mode)
{
	char hst_mode_register_offset=HST_MODE_REGISTER_OFFSET;
	char hst_mode_register_value;
	char new_hst_mode_register_value=0;

	if (verbose == VERBOSE_DEBUG)
		printk("Enter in switch to other mode, target mode=%d\n", device_mode);

	//Set offset of HOST MODE register
	if (psoc_bootloader_set_offset_register(verbose, hst_mode_register_offset))
		return 1;

	udelay(DELAY_BETWEEN_SET_OFFSET_REG_AND_RW_REG_VALUE);

	//Read value of HOST MODE register
	if (psoc_bootloader_read_register(verbose, &hst_mode_register_value))
		return 1;

	new_hst_mode_register_value = (hst_mode_register_value & ~HST_MODE_REGISTER_DEVICE_MODE_MASK ) | (device_mode << HST_MODE_REGISTER_DEVICE_MODE_OFFSET);

	if (verbose == VERBOSE_DEBUG){	
		printk("Switch to other mode, Old hst mode register value: 0x%x\n", hst_mode_register_value);
		printk("Switch to other mode, New hst mode register value: 0x%x\n", new_hst_mode_register_value);
	}

	//Write value of HOST MODE register
	if (psoc_bootloader_write_register(verbose, hst_mode_register_offset, new_hst_mode_register_value))
		return 1;

	current_mode = (int) device_mode;

	if (verbose == VERBOSE_DEBUG)
		printk("Exit switch to other mode\n");

	return 0;

}

static void psoc_bootloader_recalibrate(void)
{
	char buff_temp[3];
	int ret;

	//Switch to sys info mode
	buff_temp[0]=0x00;
	buff_temp[1]=0x10;

	ret=psoc_bootloader_i2c_send(buff_temp, 2, VERBOSE_DEBUG);
	if ( ret != 2 )
		return;

	msleep(DELAY_TO_SWITCH_TO_SYS_INFO_MODE);

	//Run mfg test command
	buff_temp[0]=0x00;
	buff_temp[1]=0x10;
	buff_temp[2]=0x30;

	ret=psoc_bootloader_i2c_send(buff_temp, 3, VERBOSE_DEBUG);
	if ( ret != 3 )
		return;

	msleep(DELAY_TO_RUN_AUTO_RECALIBRATION);

}

static int psoc_bootloader_i2c_start_ts(void)
{

	char *buff_temp;
	int size_buff_temp;
	int ret;

	printk(KERN_ERR PSOC_DEVNAME ": Start the touch screen application\n");
	size_buff_temp = SIZE_PREFIX_BOOTLOADER_CMD + SIZE_BOOTLOADER_CMD + SIZE_BOOTLOADER_KEY;
	buff_temp = (char*) kmalloc(size_buff_temp*sizeof(char), GFP_KERNEL);

	//Command prefix
	buff_temp[0]=0x00;
	buff_temp[1]=0x00;

	//Bootloader command
	buff_temp[2]=BOOTLOADER_CMD_PREFIX;
	buff_temp[3]=RUN_APP_CMD;

	//Bootloader key
	memcpy((void*) (buff_temp + 4), (void*) pdata.key, SIZE_BOOTLOADER_KEY*sizeof(char));

	ret=psoc_bootloader_i2c_send(buff_temp, size_buff_temp, VERBOSE_DEBUG);

	kfree(buff_temp);

	return (ret == size_buff_temp)?0:1;

}

static int psoc_bootloader_try_i2c_communication(void)
{
	int i=0;
	char test;

	while (i<MAX_RETRIES_BL_ACCESS){
		msleep(DELAY_BETWEEN_RELEASE_RST_AND_SEND_1ST_CMD);
		if ( psoc_bootloader_i2c_recv(&test, 1, VERBOSE_ERROR) == 1)
			break;
		i++;
	}

	if (i<MAX_RETRIES_BL_ACCESS)
		return 0;
	else
		return 1;
	
}

static void psoc_bootloader_loading_finished(const struct firmware *fw, void* context)
{

	int ret;

	printk(KERN_ERR PSOC_DEVNAME ": request firmware finished\n");
	if ( fw != NULL )
		memcpy((void*) &psoc_firmware, (void*) fw, sizeof(struct firmware));
	else
		return;

	printk(KERN_ERR PSOC_DEVNAME ": firmware received\n");
	printk(KERN_ERR PSOC_DEVNAME ": update in progress...\n");

	ret = psoc_bootloader_parse_lines(VERBOSE_LEVEL);

	msleep(DELAY_TO_REBOOT_AFTER_UPDATE);
	pdata.reset_ts();
	psoc_bootloader_try_i2c_communication();
	psoc_bootloader_i2c_start_ts();
	msleep(DELAY_TO_START_TS_APPLICATION);
	psoc_bootloader_recalibrate();

	if (ret == 0 )
		printk(KERN_ERR PSOC_DEVNAME ": update of the capacitive touch screen firmware successful ...\n");
	else
		printk(KERN_ERR PSOC_DEVNAME ": update of the capacitive touch screen firmware failed ...\n");

}

static ssize_t psoc_bootloader_fw_version_show(struct device *dev,
                                     struct device_attribute *attr, char *buf)
{
	ssize_t temp;

	if (down_interruptible(&drv_data->mutex))
		return -ERESTARTSYS;

	temp = sprintf(buf, "%X.%.2X\n", i2c_packet_received_sys_info.application_version_major, i2c_packet_received_sys_info.application_version_minor);

	up(&drv_data->mutex);

	return temp;
}

static ssize_t psoc_bootloader_fw_version_store(struct device *dev,
                                     struct device_attribute *attr,
                                     const char *buf, size_t count)
{
	return count;
}

static DEVICE_ATTR(fw_version, 0664, psoc_bootloader_fw_version_show, psoc_bootloader_fw_version_store);

static struct attribute *psoc_bootloader_attributes[] = {
	&dev_attr_fw_version.attr,
	NULL,
};

static struct attribute_group psoc_bootloader_attr_group = {
	.attrs = psoc_bootloader_attributes,
};

static void psoc_bootloader_update_work(struct work_struct *work)
{
	int ret;
	ret = request_firmware_nowait(THIS_MODULE, FW_ACTION_HOTPLUG, "fw_cap_ts.dld", &psoc_bootloader_i2c.client->dev, NULL, psoc_bootloader_loading_finished);
	if (ret != 0)
		printk(KERN_ERR PSOC_DEVNAME ": Request firmware failed, return code = %d\n", ret);
}

static void psoc_bootloader_sys_info_work(struct work_struct *work)
{

	if (down_interruptible(&drv_data->mutex))
		return;

	interrupt_triggered=1;

	if (current_mode == SYS_INFO_MODE)  {
		psoc_bootloader_set_offset_register(VERBOSE_LEVEL, HST_MODE_REGISTER_OFFSET);
		udelay(DELAY_BETWEEN_SET_OFFSET_REG_AND_RW_REG_VALUE);
		psoc_bootloader_get_system_info(VERBOSE_LEVEL);
		current_mode=OPERATING_MODE;
		disable_irq(psoc_bootloader_i2c.client->irq);
		free_irq(psoc_bootloader_i2c.client->irq, NULL);
		pdata.reset_ts();
		psoc_bootloader_try_i2c_communication();
		printk(KERN_ERR PSOC_DEVNAME ": Capacitive Touch Screen Reset\n" );

		printk(KERN_ERR PSOC_DEVNAME ": Start the update firmware application\n");
		schedule_work(&psoc_bootloader_i2c.update_work);

	}
	
	up(&drv_data->mutex);

}

static void psoc_bootloader_timeout_work(struct work_struct *work)
{

	if (down_interruptible(&drv_data->mutex))
                return;

	if (!interrupt_triggered) {
		i2c_packet_received_sys_info.application_version_major=0x0F;
        	i2c_packet_received_sys_info.application_version_minor=0xFE;
		printk(KERN_ERR PSOC_DEVNAME ": Can't read firmware version, most probably previous firmware update interrupted\n");
		printk(KERN_ERR PSOC_DEVNAME ": Now the touch screen is blanck\n" );
		current_mode=OPERATING_MODE;
		disable_irq(psoc_bootloader_i2c.client->irq);
		free_irq(psoc_bootloader_i2c.client->irq, NULL);
		pdata.reset_ts();
		psoc_bootloader_try_i2c_communication();
		printk(KERN_ERR PSOC_DEVNAME ": Capacitive Touch Screen Reset\n" );

		printk(KERN_ERR PSOC_DEVNAME ": Start the update firmware application\n");
		schedule_work(&psoc_bootloader_i2c.update_work);
	}

	up(&drv_data->mutex);

}

static void psoc_bootloader_timer(unsigned long data)
{
	schedule_work(&psoc_bootloader_i2c.timeout_work);
}

static irqreturn_t psoc_bootloader_irq_handler(int irq, void *dev_id)
{
	schedule_work(&psoc_bootloader_i2c.sys_info_work);
	return IRQ_HANDLED;
};

static void psoc_bootloader_i2c_init(struct i2c_client *client)
{
	BUG_ON(!client);
	//Init psoc_i2c struct
	memset(&psoc_bootloader_i2c, 0x00, sizeof(psoc_bootloader_i2c));
	psoc_bootloader_i2c.client = client;
	memcpy((void*) &pdata, (void*) client->dev.platform_data, sizeof(psoc_pdata_t));
	INIT_WORK(&psoc_bootloader_i2c.update_work, psoc_bootloader_update_work);
	INIT_WORK(&psoc_bootloader_i2c.sys_info_work, psoc_bootloader_sys_info_work);
	INIT_WORK(&psoc_bootloader_i2c.timeout_work, psoc_bootloader_timeout_work);
	drv_data = kmalloc(sizeof(psoc_drv_data_t), GFP_KERNEL);

	if(NULL == drv_data) {
		printk(KERN_ERR PSOC_DEVNAME ": Can allocate driver data!\n");
		return;
	}

	dev_set_drvdata(&client->dev, drv_data);
	init_MUTEX(&drv_data->mutex);

	//Init Timer
	init_timer(&mytimer);
	mytimer.function=&psoc_bootloader_timer;
	mytimer.expires = jiffies + msecs_to_jiffies(2000);
}


static int psoc_bootloader_i2c_probe(struct i2c_client *client, const struct i2c_device_id *id)
{

	char tmp;
	int err;

	len_i2c_frame_sys_info = sizeof(struct i2c_data_received_sys_info);
	i2c_packet_received_sys_info.application_version_major=0;
	i2c_packet_received_sys_info.application_version_minor=0;

	index_firmware=0;
	interrupt_triggered=0;

	printk(KERN_INFO PSOC_DEVNAME ": probing driver\n");

	printk(KERN_ERR PSOC_DEVNAME ": Psoc Bootloader Capacitive Touch Screen Driver, (C) 2008 TomTom BV\n" );

	BUG_ON(!client);
	BUG_ON(!client->dev.platform_data);
	psoc_bootloader_i2c_init(client);

	pdata.reset_ts();

	if (!psoc_bootloader_try_i2c_communication())
	{

		psoc_bootloader_i2c_start_ts();
		msleep(DELAY_TO_START_TS_APPLICATION);

		//If it is possible to read ts app it means that the fw version is
		//3.32 or above so it is possible to read the fw version
		//Otherwise version 0.00 is returned in sysfs
		if (!psoc_bootloader_read_register(VERBOSE_DISABLED, &tmp))
		{

			printk(KERN_INFO PSOC_DEVNAME ": irq number [%d]\n", client->irq);

			set_irq_type(client->irq, IRQ_TYPE_EDGE_RISING);
			if (request_irq(client->irq, &psoc_bootloader_irq_handler, IRQ_TYPE_EDGE_FALLING, "psoc_1", NULL)) {
				printk(KERN_ERR PSOC_DEVNAME " Could not allocate IRQ (psoc_1)!\n");
				kfree(drv_data);
				return -EIO;
			}

			if (down_interruptible(&drv_data->mutex))
				return -ERESTARTSYS;

			//Switch to system information mode
			psoc_bootloader_switch_to_other_mode(VERBOSE_LEVEL, SYS_INFO_MODE);

			up(&drv_data->mutex);

			add_timer(&mytimer);

		}
		else
		{
			pdata.reset_ts();
                	psoc_bootloader_try_i2c_communication();
                	printk(KERN_ERR PSOC_DEVNAME ": Capacitive Touch Screen Reset\n" );	
			schedule_work(&psoc_bootloader_i2c.update_work);
		}
	}
	else
	{
		i2c_packet_received_sys_info.application_version_major=0x0F;
        	i2c_packet_received_sys_info.application_version_minor=0xFF;
		printk(KERN_ERR PSOC_DEVNAME ": I2C Communication Impossible with the Bootloader\n" );
		printk(KERN_ERR PSOC_DEVNAME ": Most Probably Wrong Factory Data\n" );
		printk(KERN_ERR PSOC_DEVNAME ": Check the Field /features/ts\n" );
		printk(KERN_ERR PSOC_DEVNAME ": Change the Value 300E into 300E_FW_V2\n" );
		printk(KERN_ERR PSOC_DEVNAME ": Or the Other Way Around\n" );
	}

	err = sysfs_create_group(&client->dev.kobj, &psoc_bootloader_attr_group);
	if (err)
		return err;

	return 0;
}

static int psoc_bootloader_i2c_remove(struct i2c_client *client)
{
	printk(KERN_INFO PSOC_DEVNAME ": start remove\n");
	sysfs_remove_group(&client->dev.kobj, &psoc_bootloader_attr_group);
	kfree(drv_data);
	return 0;
}

static int psoc_bootloader_i2c_suspend(struct i2c_client *client, pm_message_t mesg)
{
	printk(KERN_INFO PSOC_DEVNAME ": start suspending\n");
	return 0;
}

static int psoc_bootloader_i2c_resume(struct i2c_client *client)
{
	printk(KERN_INFO PSOC_DEVNAME ": start resuming\n");
	return 0;
}

static const struct i2c_device_id psoc_bootloader_id[] = {
	{ PSOC_DEVNAME, 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, psoc_bootloader_id);

static struct i2c_driver psoc_bootloader_i2c_driver=
{
	.probe	= psoc_bootloader_i2c_probe,
	.remove	= psoc_bootloader_i2c_remove,
	.suspend= psoc_bootloader_i2c_suspend,
	.resume	= psoc_bootloader_i2c_resume,
	.id_table = psoc_bootloader_id,
	.driver = {
		.name	= PSOC_DEVNAME,
		.owner	= THIS_MODULE,
	},
};

static int __init psoc_bootloader_init(void)
{
	int err;

	printk(KERN_ERR PSOC_DEVNAME ": start init\n");

	if ((err = i2c_add_driver(&psoc_bootloader_i2c_driver))) {
		printk(KERN_ERR PSOC_DEVNAME ": Could Not Be Added. Err Code: [%i]\n", err);
	}

	return err;
}

static void __exit psoc_bootloader_exit(void)
{
	printk(KERN_INFO PSOC_DEVNAME ": start exit\n");
	i2c_del_driver(&psoc_bootloader_i2c_driver);
}

MODULE_AUTHOR("Vincent Dejouy <vincent.dejouy@tomtom.com>");
MODULE_DESCRIPTION("Driver for Bootloader Of PSoc, Capacitive Touch Screen Controller.");
MODULE_LICENSE("GPL");

module_init(psoc_bootloader_init);
module_exit(psoc_bootloader_exit);

