/* psoc-ctsic.c
 *
 * Control driver for PSOC Capacitive TouchScreen IC
 *
 * Copyright (C) 2006 TomTom BV <http://www.tomtom.com/>
 * Authors: Vincent Dejouy <vincent.dejouy@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/cdev.h>
#include <linux/i2c.h>
#include <linux/workqueue.h>
#include <linux/spinlock.h>
#include <linux/err.h>
#include <linux/psoc-ctsic.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/timer.h>
#include <linux/jiffies.h>


#define NUM_MAX_FINGERS 2
#define FINGERS_EXPORTED 2

#define VERBOSE_DEBUG    2
#define VERBOSE_ERROR    1
#define VERBOSE_DISABLED 0

#define VERBOSE_LEVEL    VERBOSE_ERROR

//Limits of the values sent to tslib or HAL
#define XMIN 		0
#define XMAX 		0xDC
#define YMIN 		0
#define YMAX		0x140
#define PRESSUREMIN	0
#define PRESSUREMAX	1
#define GESTUREMIN	0
#define GESTUREMAX	0x10
#define FINGERSMIN	0
#define FINGERSMAX	2

struct input_dev *vms_input_dev;     /* Representation of an input device */

typedef struct
{
	struct semaphore		mutex;
} psoc_drv_data_t;

struct _psoc_i2c {
	struct i2c_client	*client;
	struct work_struct	work;
} psoc_i2c;

struct finger_position
{
	unsigned char finger_id;
	unsigned short age;
	unsigned char z;
	unsigned short y;
	unsigned short x;
} __attribute__((__packed__));

struct i2c_data_received
{
	unsigned char type;
	unsigned char size;
	struct finger_position finger_positions[NUM_MAX_FINGERS];
	unsigned char fingers;
} __attribute__((__packed__));

static struct i2c_data_received i2c_packet_received;

//Uncomment to use finger id tracking mechanism
/*typedef struct
{
	int x;
	int y;
} finger_position_to_export_t;

static finger_position_to_export_t fingers[FINGERS_EXPORTED];
static unsigned char finger_ids[FINGERS_EXPORTED]={0xFF,0xFF};*/

static struct timer_list mytimer;


//Uncomment to use finger id tracking mechanism
/*static void psoc_update_finger_positions(void)
{
	int i,j;

	i=0;
	while ( (finger_ids[i] != 0xFF) && (i < FINGERS_EXPORTED))
	{
		for ( j=0; j<NUM_MAX_FINGERS; j++ ) {
			if ( i2c_packet_received.finger_positions[j].finger_id == finger_ids[i] ) {
				fingers[i].x = (int) i2c_packet_received.finger_positions[j].x;
				fingers[i].y = (int) i2c_packet_received.finger_positions[j].y;
				break;
			}
		}
		i++;
	}
}*/

static inline unsigned short psoc_convert_to_raw_value(char* number_to_convert)
{
	return ( (number_to_convert[1]) | ( number_to_convert[0] << 8) );
}

/**
 * psoc_i2c_send - issue a single I2C message in master transmit mode
 * @buf: Data that will be written to the slave
 * @count: How many bytes to write
 *
 * Returns negative errno, or else the number of bytes written.
 */
static int psoc_i2c_send(const char *buf, int count)
{
	int ret = -ENODEV;
	psoc_drv_data_t *drv_data = dev_get_drvdata(&psoc_i2c.client->dev);
	BUG_ON(!buf);
	if (down_interruptible(&drv_data->mutex))
		return -ERESTARTSYS;
	ret = i2c_master_send(psoc_i2c.client, buf, count);
	up(&drv_data->mutex);
	return ret;
}


/**
 * psoc_i2c_recv - issue a single I2C message in master receive mode
 * @buf: Where to store data read from slave
 * @count: How many bytes to read
 *
 * Returns negative errno, or else the number of bytes read.
 */
static int psoc_i2c_recv(char *buf, int count)
{
	int ret = -ENODEV;
	psoc_drv_data_t *drv_data = dev_get_drvdata(&psoc_i2c.client->dev);
	BUG_ON(!buf);
	if (down_interruptible(&drv_data->mutex))
		return -ERESTARTSYS;
	ret = i2c_master_recv(psoc_i2c.client, buf, count);
	up(&drv_data->mutex);
	return ret;
}

static int psoc_get_finger_position(int verbose)
{

        int ret;
	int i;
	int len_i2c_frame;
	char* buff_temp;
	unsigned short temp;

	len_i2c_frame = sizeof(struct i2c_data_received);

	buff_temp = (char*) kmalloc(len_i2c_frame, GFP_KERNEL);

        ret = psoc_i2c_recv(buff_temp, len_i2c_frame);

	
        if (ret != len_i2c_frame){
		if (verbose)
                	printk(KERN_ERR PSOC_301_DEVNAME ": Error reading finger positions, return value of i2c read: [%d] instead of [%d]\n", ret, len_i2c_frame + 5);
                kfree(buff_temp);
                return 1;
        }


	if (verbose == VERBOSE_DEBUG){
		printk(PSOC_301_DEVNAME ":data received:\n");
		for (i=0;i<len_i2c_frame;i++){
			printk("0x%x ",buff_temp[i]);
		}
		printk("\n");
	}

	memcpy((void*) &i2c_packet_received, (void*) buff_temp, len_i2c_frame);
	

	if (verbose == VERBOSE_DEBUG){
		printk(PSOC_301_DEVNAME ":i2c packet received before finger positions reversed:\n");
		printk("Type: 0x%x\n", i2c_packet_received.type);
		printk("Size: 0x%x\n", i2c_packet_received.size);

		for (i=0; i<NUM_MAX_FINGERS;i++) {
			printk("\nFinger %d:\n", i);
			printk("Finger ID: 	%d\n", i2c_packet_received.finger_positions[i].finger_id);
			printk("Age: 		%d\n", i2c_packet_received.finger_positions[i].age);
			printk("X: 		0x%x\n", i2c_packet_received.finger_positions[i].x);
			printk("Y: 		0x%x\n", i2c_packet_received.finger_positions[i].y);
			printk("Z: 		0x%x\n", i2c_packet_received.finger_positions[i].z);
		}

		printk("\nFingers: %d\n", i2c_packet_received.fingers);
		printk("\n");
	}

	for (i=0; i<NUM_MAX_FINGERS;i++) {
		temp = psoc_convert_to_raw_value((char*) &i2c_packet_received.finger_positions[i].age);
		i2c_packet_received.finger_positions[i].age = temp;
		temp = psoc_convert_to_raw_value((char*) &i2c_packet_received.finger_positions[i].x);
		i2c_packet_received.finger_positions[i].x = temp;
		temp = psoc_convert_to_raw_value((char*) &i2c_packet_received.finger_positions[i].y);
		i2c_packet_received.finger_positions[i].y = temp;
	}

	if (verbose == VERBOSE_DEBUG){
                printk(PSOC_301_DEVNAME ":i2c packet received after finger positions reversed:\n");
                printk("Type: 0x%x\n", i2c_packet_received.type);
                printk("Size: 0x%x\n", i2c_packet_received.size);

                for (i=0; i<NUM_MAX_FINGERS;i++) {
                        printk("\nFinger %d:\n", i);
                        printk("Finger ID:      %d\n", i2c_packet_received.finger_positions[i].finger_id);
                        printk("Age:            %d\n", i2c_packet_received.finger_positions[i].age);
                        printk("X:              0x%x\n", i2c_packet_received.finger_positions[i].x);
                        printk("Y:              0x%x\n", i2c_packet_received.finger_positions[i].y);
                        printk("Z:              0x%x\n", i2c_packet_received.finger_positions[i].z);
                }

                printk("\nFingers: %d\n", i2c_packet_received.fingers);
                printk("\n");
        }

	//Uncomment to use finger id tracking mechanism
	/*switch (i2c_packet_received.fingers) {
		case 0: for (i=0; i<FINGERS_EXPORTED;i++)
				finger_ids[i]=0xFF;
			break;
		case 1: finger_ids[0]=i2c_packet_received.finger_positions[0].finger_id;
			finger_ids[1]=0xFF;
			break;
		case 2:
		case 3:
		case 4: if ( finger_ids[0] == 0xFF ) {
				finger_ids[0]=i2c_packet_received.finger_positions[0].finger_id;
				finger_ids[1]=i2c_packet_received.finger_positions[1].finger_id;
			} else {
				if ( finger_ids[1] == 0xFF ) {
					if ( finger_ids[0] == i2c_packet_received.finger_positions[0].finger_id )
						finger_ids[1] = i2c_packet_received.finger_positions[1].finger_id;
					else
						finger_ids[1] = i2c_packet_received.finger_positions[0].finger_id;
				}
			}
			break;
		default: printk(KERN_ERR PSOC_301_DEVNAME ": Error too many fingers on the screen, only 4 supported\n");
			 kfree(buff_temp);
			 return;
	}

	psoc_update_finger_positions();

	if ( i2c_packet_received.fingers > FINGERS_EXPORTED )
		i2c_packet_received.fingers = FINGERS_EXPORTED;*/

        //Report position of finger to the input layer
        /*input_report_abs(vms_input_dev, ABS_X, fingers[0].x);
        input_report_abs(vms_input_dev, ABS_Y, fingers[0].y);
	input_report_abs(vms_input_dev, ABS_FINGER2_X, fingers[1].x);
	input_report_abs(vms_input_dev, ABS_FINGER2_Y, fingers[1].y);*/

	input_report_abs(vms_input_dev, ABS_FINGERS, i2c_packet_received.fingers);

	input_report_abs(vms_input_dev, ABS_X, i2c_packet_received.finger_positions[0].x);
        input_report_abs(vms_input_dev, ABS_Y, i2c_packet_received.finger_positions[0].y);
        input_report_abs(vms_input_dev, ABS_FINGER2_X, i2c_packet_received.finger_positions[1].x);
        input_report_abs(vms_input_dev, ABS_FINGER2_Y, i2c_packet_received.finger_positions[1].y);

	if (i2c_packet_received.fingers)
		input_report_abs(vms_input_dev, ABS_PRESSURE, 1);
	else
		input_report_abs(vms_input_dev, ABS_PRESSURE, 0);

        input_sync(vms_input_dev);
        kfree(buff_temp);

	return 0;

}

static int psoc_clear_busy_flag(int verbose)
{
        int ret;
        char buff_clearance[2];
	int i;

        buff_clearance[0]=1;
        buff_clearance[1]=4;
     
        ret = psoc_i2c_send(buff_clearance, 2);

	if (verbose == VERBOSE_DEBUG){
		printk(PSOC_301_DEVNAME ":data sent:\n");
		for (i=0;i<2;i++){
                        printk("0x%x ",buff_clearance[i]);
                }
                printk("\n");
	}

        if (ret != 2 && verbose)
                printk(KERN_ERR PSOC_301_DEVNAME ": Error clearing busy flag, return value of i2c send: [%d] instead of 2\n", ret);
          
	return (ret == 2)?0:1;
    
}

static void psoc_work(struct work_struct *work)
{
        psoc_clear_busy_flag(VERBOSE_LEVEL);
        psoc_get_finger_position(VERBOSE_LEVEL);
}


// Timer handler
static void psoc_timer(unsigned long data)
{ 
        schedule_work(&psoc_i2c.work);
        mytimer.expires = jiffies + msecs_to_jiffies(50);
        add_timer(&mytimer);
}

static irqreturn_t psoc_irq_handler(int irq, void *dev_id)
{
        schedule_work(&psoc_i2c.work);
        return IRQ_HANDLED;
};

static void psoc_i2c_init(struct i2c_client *client)
{
        
	BUG_ON(!client);
        
        //Init psoc_i2c struct
	memset(&psoc_i2c, 0x00, sizeof(psoc_i2c));
        psoc_i2c.client = client;
	INIT_WORK(&psoc_i2c.work, psoc_work);

        //Init Timer
        init_timer(&mytimer);
        mytimer.function=&psoc_timer;
        mytimer.expires = jiffies + msecs_to_jiffies(50);

}


static int psoc_i2c_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	psoc_drv_data_t *drv_data;
	psoc_pdata_t	*pdata;
	int ret, i;

        drv_data = NULL;
        pdata = (psoc_pdata_t *)client->dev.platform_data;

	if(NULL == pdata) {
		printk(KERN_ERR PSOC_301_DEVNAME  ": Missing platform data!\n");
		return -ENODEV;
	}

	BUG_ON(!client);
	psoc_i2c_init(client);

	drv_data = kmalloc(sizeof(psoc_drv_data_t), GFP_KERNEL);

	if(NULL == drv_data) {
		printk(KERN_ERR PSOC_301_DEVNAME ": Can allocate driver data!\n");
		return -ENOMEM;
	}

	dev_set_drvdata(&client->dev, drv_data);
	init_MUTEX(&drv_data->mutex);
        vms_input_dev = input_allocate_device();

        if (!vms_input_dev) {
                printk(KERN_ERR PSOC_301_DEVNAME ": Badf init_alloc_device()\n");
                kfree(drv_data);
                return -10;
        }

        printk(KERN_INFO PSOC_301_DEVNAME ": Input device allocated\n");

        /* Announce that the ts will generate absolute coordinates */
        set_bit(EV_ABS, vms_input_dev->evbit);
        printk(KERN_INFO PSOC_301_DEVNAME ": Event Abs positions declared\n");

	
        // Declare max and min of X, Y and pressure
        input_set_abs_params(vms_input_dev, ABS_X, XMIN, XMAX, 0, 0);
        input_set_abs_params(vms_input_dev, ABS_Y, YMIN, YMAX, 0, 0);
        input_set_abs_params(vms_input_dev, ABS_PRESSURE, PRESSUREMIN, PRESSUREMAX, 0, 0);

	input_set_abs_params(vms_input_dev, ABS_FINGERS, FINGERSMIN, FINGERSMAX, 0, 0);
	input_set_abs_params(vms_input_dev, ABS_FINGER2_X, XMIN, XMAX, 0, 0);
        input_set_abs_params(vms_input_dev, ABS_FINGER2_Y, YMIN, YMAX, 0, 0);
        input_set_abs_params(vms_input_dev, ABS_GESTURE, GESTUREMIN, GESTUREMAX, 0, 0);

        vms_input_dev->name = "PSOC Cypress Capacitive TouchScreen";
        vms_input_dev->phys = "input(ts)";
        vms_input_dev->id.vendor = 0xDEAD;
        vms_input_dev->id.product = 0xBEEF;
        vms_input_dev->id.version = 0x0101;

        /* Register with the input subsystem */
        ret = input_register_device(vms_input_dev);

        if (ret) {
                printk(KERN_ERR PSOC_301_DEVNAME ": Could not register input device(capacitive touchscreen)!\n");
                kfree(drv_data);
                return -EIO;
        }

        printk(KERN_INFO PSOC_301_DEVNAME ": Input device registered\n");

        set_irq_type(client->irq, IRQ_TYPE_EDGE_RISING);
        if (request_irq(client->irq, &psoc_irq_handler, IRQ_TYPE_EDGE_RISING, "psoc_1", NULL)) {
                printk(KERN_ERR PSOC_301_DEVNAME " Could not allocate IRQ [%d]\n", client->irq);
                input_unregister_device(vms_input_dev);
                kfree(drv_data);
                return -EIO;
        }

	i=0;
	while(psoc_get_finger_position(VERBOSE_DISABLED) && i<10){
		msleep(500);
		i++;
	}
	
	if (i==10)
		printk(KERN_INFO PSOC_301_DEVNAME ":Warning, can't read finger positions after 10 trials, most probably touch screen not working\n"); 

        psoc_clear_busy_flag(VERBOSE_LEVEL);

	printk (KERN_INFO "Cypress Touchscreen driver, (C) 2009 TomTom BV\n");
	printk (KERN_INFO PSOC_301_DEVNAME ": addr: 0x%x, irq: 0x%x)\n", client->addr, client->irq);

	//Uncomment to use with timer 
        //add_timer(&mytimer);

	return 0;
}

static int psoc_i2c_remove(struct i2c_client *client)
{
	psoc_drv_data_t *drv_data;

	printk(KERN_INFO PSOC_301_DEVNAME ": start remove\n");
	drv_data = dev_get_drvdata(&client->dev);
        BUG_ON(!drv_data);

        disable_irq(client->irq);
        free_irq(client->irq, NULL);

        input_unregister_device(vms_input_dev);
	kfree(drv_data);

	return 0;
}

#define psoc_i2c_suspend		NULL
#define psoc_i2c_resume			NULL

static const struct i2c_device_id psoc_id[] = {
	{ PSOC_301_DEVNAME, 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, psoc_id);

static struct i2c_driver psoc_i2c_driver=
{
	.id	= I2C_DRIVERID_PSOC_CTSIC,
	.probe	= psoc_i2c_probe,
	.remove	= psoc_i2c_remove,
	.suspend= psoc_i2c_suspend,
	.resume	= psoc_i2c_resume,
	.id_table = psoc_id,
	.driver = {
		.name	= PSOC_301_DEVNAME,
		.owner	= THIS_MODULE,
	},
};

static int __init psoc_init(void)
{
	int err;

	printk(KERN_INFO PSOC_301_DEVNAME " :start init\n");

	if ((err = i2c_add_driver(&psoc_i2c_driver))) {
		printk(KERN_ERR PSOC_301_DEVNAME ": Could Not Be Added. Err Code: [%i]\n", err);
	}

	return err;
}

static void __exit psoc_exit(void)
{
	printk(KERN_INFO PSOC_301_DEVNAME " :start exit\n");
	i2c_del_driver(&psoc_i2c_driver);
}


MODULE_AUTHOR("Vincent Dejouy <vincent.dejouy@tomtom.com>");
MODULE_DESCRIPTION("Driver for I2C connected PSoc Capacitive Touchscreen IC.");
MODULE_LICENSE("GPL");

module_init(psoc_init);
module_exit(psoc_exit);
