/*
 *  linux/drivers/mfd/s5m8751_pmic.c
 *
 *  Copyright (C) 2009 TomTom BV <http://www.tomtom.com/>
 *  Author: Andrzej Zukowski <andrzej.zukowski@tomtom.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/err.h>
#include <linux/mfd/s5m8751/s5m8751_core.h>

#include <mach/gpio.h>
#include <plat/usbmode.h>

/*
 * Conversion macros
 */
#define S5M8751_GET_ONOFF2(s) \
		((s & S5M8751_LDO1_EN ) ? S5M8751_LDO1_ENA	: 0) |\
		((s & S5M8751_LDO2_EN ) ? S5M8751_LDO2_ENA	: 0) |\
		((s & S5M8751_LDO3_EN ) ? S5M8751_LDO3_ENA	: 0) |\
		((s & S5M8751_LDO4_EN ) ? S5M8751_LDO4_ENA	: 0) |\
		((s & S5M8751_LDO5_EN ) ? S5M8751_LDO_MEMORY_ENA: 0)
#define S5M8751_GET_ONOFF3(s) \
		((s & S5M8751_LDO6_EN ) ? S5M8751_LDO_AUDIO_ENA	: 0) |\
		((s & S5M8751_BCK1_EN ) ? S5M8751_BUCK1_ENA	: 0) |\
		((s & S5M8751_BCK2_EN ) ? S5M8751_BUCK2_ENA	: 0)
#define S5M8751_GET_SLEEP_CNTL1(s) \
		((s & S5M8751_LDO1_EN ) ? S5M8751_LDO1_ENA	: 0) |\
		((s & S5M8751_LDO2_EN ) ? S5M8751_LDO2_ENA	: 0) |\
		((s & S5M8751_LDO3_EN ) ? S5M8751_LDO3_ENA	: 0) |\
		((s & S5M8751_LDO4_EN ) ? S5M8751_LDO4_ENA	: 0) |\
		((s & S5M8751_LDO5_EN ) ? S5M8751_LDO_MEMORY_ENA: 0)
#define S5M8751_GET_SLEEP_CNTL2(s) \
		((s & S5M8751_LDO6_EN ) ? S5M8751_LDO_AUDIO_ENA	: 0) |\
		((s & S5M8751_BCK1_EN ) ? S5M8751_BUCK1_ENA	: 0) |\
		((s & S5M8751_BCK2_EN ) ? S5M8751_BUCK2_ENA	: 0)

static struct s5m8751 *s5m8751;

static void s5m8751_pmic_set_sleep_state(uint32_t state)
{
	uint8_t reg_sleep_cntl1	= S5M8751_GET_SLEEP_CNTL1(state);
	uint8_t reg_sleep_cntl2	= S5M8751_GET_SLEEP_CNTL2(state);
	uint8_t reg_onoff1	= S5M8751_SLEEPB_PIN_EN;

	s5m8751_reg_write(s5m8751, S5M8751_SLEEP_CNTL1, reg_sleep_cntl1);
	s5m8751_reg_write(s5m8751, S5M8751_SLEEP_CNTL2, reg_sleep_cntl2);
	s5m8751_reg_write(s5m8751, S5M8751_ONOFF1, reg_onoff1);
}

static void s5m8751_pmic_set_wall_current(uint8_t wall_pwr_current)
{
	uint8_t reg_chg_iv_set = s5m8751_reg_read(s5m8751, S5M8751_CHG_IV_SET);

	reg_chg_iv_set &= ~S5M8751_FCHG_CSET_MASK;
	reg_chg_iv_set |= wall_pwr_current;
	s5m8751_reg_write(s5m8751, S5M8751_CHG_IV_SET, reg_chg_iv_set);
}

static void s5m8751_pmic_state_listener(USB_STATE prev_state, USB_STATE curr_state, void *arg)
{
	struct platform_s5m8751_pmic_data *pdata = (struct platform_s5m8751_pmic_data *) arg;

	if (curr_state == prev_state)
		return;

	switch (curr_state)
	{
		case USB_STATE_IDLE :
		default:
			/* nothing connected, or in transition between states, disable charging */
			printk("S5M8751: USB_STATE: [IDLE|DEFAULT]\n");
			gpio_direction_output(pdata->wall_pwr_pin, 0);
			break;
		case USB_STATE_CLA:
			/* CLA or wall pwr connected, enable 800mA charging. */
			printk("S5M8751: USB_STATE: [CLA]\n");
			gpio_direction_output(pdata->wall_pwr_pin, 1);
			break;
		case USB_STATE_DEVICE :
		case USB_STATE_DEVICE_WAIT :
		case USB_STATE_DEVICE_DETECT:
			/* something connected, enable 450mA charging. */
			printk("S5M8751: USB_STATE: [DEVICE|DEVICE_WAIT|DEVICE_DETECT]\n");
			gpio_direction_output(pdata->wall_pwr_pin, 0);
			break;
	}
}

static int s5m8751_pmic_probe(struct platform_device *pdev)
{
	struct platform_s5m8751_pmic_data *pdata = pdev->dev.platform_data;
	USB_STATE curr_state;
	int ret;

	s5m8751 = dev_to_s5m8751(&pdev->dev);

	s5m8751_pmic_set_sleep_state(pdata->sleep_mode_state);
	s5m8751_pmic_set_wall_current(pdata->wall_pwr_current);

	if (add_usb_state_change_listener(s5m8751_pmic_state_listener, pdata, &curr_state) != 0)
	{
		printk( KERN_ERR "S5M8751: Can't register USBMODE change state listener.\n" );
		ret = -ENODEV;
		goto err_usb;
	}

	if (pdata->request_gpio)
	{
		if ((ret = pdata->request_gpio()))
		{
			printk(KERN_ERR "S5M8751: Can't initialize GPIOs.\n");
			goto err_gpio;
		}

		if (pdata->config_gpio)
			pdata->config_gpio();
	}

	s5m8751_pmic_state_listener(USB_STATE_INITIAL, curr_state, pdata);
	return 0;

err_gpio:
	remove_usb_state_change_listener( s5m8751_pmic_state_listener, NULL);

err_usb:
	return ret;
}

static int s5m8751_pmic_remove(struct platform_device *pdev)
{
	struct platform_s5m8751_pmic_data *pdata = pdev->dev.platform_data;

	if (pdata->free_gpio)
		pdata->free_gpio();

	if (remove_usb_state_change_listener( s5m8751_pmic_state_listener, NULL) != 0)
		printk("S5M8751: Couldn't remove usb state change listener!\n");

	return 0;
}

#ifdef CONFIG_PM
static int s5m8751_pmic_suspend(struct platform_device *pdev, pm_message_t state)
{
	if (remove_usb_state_change_listener( s5m8751_pmic_state_listener, NULL) != 0)
		printk("S5M8751: Couldn't remove usb state change listener!\n");

	return 0;
}

static int s5m8751_pmic_resume(struct platform_device *pdev)
{
	struct platform_s5m8751_pmic_data *pdata = pdev->dev.platform_data;
	USB_STATE curr_state;

	if (add_usb_state_change_listener(s5m8751_pmic_state_listener, pdata, &curr_state) != 0)
	{
		printk(KERN_ERR "S5M8751: Can't register USBMODE change state listener.\n");
		return -ENODEV;
	}

	s5m8751_pmic_state_listener(USB_STATE_INITIAL, curr_state, pdata);
	return 0;
}
#else
#define s5m8751_pmic_suspend	NULL
#define s5m8751_pmic_resume	NULL
#endif

static struct platform_driver s5m8751_pmic_driver = {
	.driver		= {
		.name	= "s5m8751-pmic",
		.owner	= THIS_MODULE,
	},
	.probe		= s5m8751_pmic_probe,
	.remove		= s5m8751_pmic_remove,
	.suspend	= s5m8751_pmic_suspend,
	.resume		= s5m8751_pmic_resume,
};

static int __init s5m8751_pmic_init(void)
{
	return platform_driver_register(&s5m8751_pmic_driver);
}
module_init(s5m8751_pmic_init);

static void __exit s5m8751_pmic_exit(void)
{
	platform_driver_unregister(&s5m8751_pmic_driver);
}
module_exit(s5m8751_pmic_exit);

MODULE_DESCRIPTION("s5m8751 based PMIC Driver");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andrzej Zukowski <andrzej.zukowski@tomtom.com>");

