/*****************************************************************************
 * Copyright 2003 - 2010 Broadcom Corporation.  All rights reserved.
 *
 * Unless you and Broadcom execute a separate written software license
 * agreement governing use of this software, this software is licensed to you
 * under the terms of the GNU General Public License version 2, available at
 * http://www.broadcom.com/licenses/GPLv2.php (the "GPL").
 *
 * Notwithstanding the above, under no circumstances may you combine this
 * software in any way with any other Broadcom software provided under a
* license other than the GPL, without Broadcom's express prior written
 * consent.
 *****************************************************************************/

/** BCM59040 Core Driver Module
 *****************************************************************************
 *
 *     This implements the BCM59040 device core functionality. The core
 *     includes the IRQChip "emulation" layer, the PMU Bus backend driver
 *     and the generic IOCTL layer.
 *
 *     The IRQChip emulation layer translates BCM59040 specific interrupt
 *     events identified from the interrupt registers into individual
 *     IRQ events that are passed onwards as different IRQChip events. Other
 *     devices register standard Linux ISRs for handling these events. It
 *     also ensures that PMU interrupts are properly configured for the
 *     suspend state. This means that any interrupt exported by the IRQChip
 *     emulation layer that is set as a wakeup event remains enabled during
 *     suspend while any interrupt that is not set as a wakeup event is
 *     masked while in suspend. The prior state of the interrupt masking
 *     is preserved and correctly restored during the resume process.
 *
 *     The PMU Bus backend device translates the underlying I2C communication
 *     required for the BCM59040 into a more generic interface. This allows
 *     abstraction from the specific communications method used by a specific
 *     PMU device.
 *
 *     The generic IOCTL layer takes all input IOCTL calls and routes them
 *     to a set of registered child devices based upon the range of IOCTL
 *     values claimed by a particular child device. The child devices make
 *     calls to this core driver to request a range of IOCTL values for
 *     themselves. If the range overlaps the values already claimed by
 *     another child driver then the request is refused, otherwise it
 *     is accepted by the core driver and registered to that child device.
 *     When an IOCTL is received that has a value within the range claimed
 *     by a child device the IOCTL is forward to that child device for
 *     handling. If the IOCTl value is unknown (unclaimed) and standard
 *     error code is returned.
 *
 * Copyright (c) 2009-2010 TomTom B.V
 * Copyright (c) 2009-2010 Broadcom Corporation
 *
 ****************************************************************************/

/* ---- Include Files ---------------------------------------------------- */

#include <linux/version.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/sysctl.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/platform_device.h>
#include <linux/kthread.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/gpio.h>
#include <linux/delay.h>
#ifdef CONFIG_REGULATOR
#include <linux/regulator/machine.h>
#endif

#include <linux/pmu_device.h>
#include <asm/arch/pmu_device_bcm59040.h>

/* Include external interfaces */

#include <linux/broadcom/pmu_bcm59040.h>
#include <asm/arch/bcm59040_irqs.h>

/*
 * Structure and macro definitions.
 */

#define BCM59040_CORE_MOD_DESCRIPTION	"BCM59040 Core Driver"
#define BCM59040_CORE_MOD_VERSION		1.0

#undef BCM59040_DEVNAME
#define BCM59040_DEVNAME	"bcm59040"
#define PFX			BCM59040_DEVNAME ": "

#define BCM59040_I2C_ADDR_IND	0
#define BCM59040_I2C_DATA_IND	1

#define BCM59040_I2C_MSG_LEN	2

#define BCM59040_PAGE_MAX_ADDR 0xff

#define BCM59040_VIRQ_NAME_MAX_LEN 64

typedef enum {EPAGE_0, EPAGE_1} e_page;
typedef enum {EIRQ_UNMASK, EIRQ_MASK} e_irq_mask;

typedef struct _bcm59040_ctxt
{
	struct i2c_client *	client;
	unsigned char		page;
	struct mutex		mutex_i2c_rw;
	struct semaphore	sem_isr;
	int			irq;
	struct task_struct *	ktask;
	unsigned char           int_regs[BCM59040_NUM_INT_REG];
	unsigned char           masks[BCM59040_NUM_INT_REG];
	u8			masked[BCM59040_NUM_INT_REG];
	int			saved;
	struct pmu_adapter		adapter;
} bcm59040_ctxt_t;

typedef struct virtual_int_registers
{
	u16		reg_addr;		/* register address */
	u16		mask_reg_addr;	/* mask register address */
	u16		reg_mask;		/* mask value for register */
} virtual_int_registers_t;

typedef struct virtual_int_entry
{
	int			int_vector;		/* vector number of interrupt */
	int			num_registers;		/* number of registers in register list */
	virtual_int_registers_t	*reg_list;		/* List of associated interrupt registers */
	int			wakeup;			/* 0 if not a wakeup source, 1 if it it is. */
} virtual_int_entry_t;

/* Debug logging */
#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG 1

#define DBG_ERROR	0x01
#define DBG_INFO	0x02
#define DBG_TRACE	0x04
#define DBG_TRACE2	0x08
#define DBG_DATA	0x10
#define DBG_DATA2	0x20

#ifdef DEBUG
//#define DBG_DEFAULT_LEVEL	(DBG_ERROR | DBG_INFO | DBG_TRACE)
#define DBG_DEFAULT_LEVEL	(DBG_ERROR)


static int gLevel = DBG_DEFAULT_LEVEL;

#	define PMU_DEBUG(level,x) {if (level & gLevel) printk x;}
#else
#	define PMU_DEBUG(level,x)
#endif

static int bcm59040_bus_read(struct pmu_adapter *adap, int address, void *data, int len);
static int bcm59040_bus_write(struct pmu_adapter *adap, int address, void *data, int len);

/*
 * Local variables.
 */

virtual_int_registers_t bcm59040_vint_table_ponkey[] = {
	{ BCM59040_REG_VINT1, BCM59040_REG_INT1M, (BCM59040_INT1_PONKEYR |
											   BCM59040_INT1_PONKEYF |
											   BCM59040_INT1_PONKEYH |
											   BCM59040_INT1_PONKEYBHD |
											   BCM59040_INT1_RESTARTH |
											   BCM59040_INT1_RESTARTON) }
};

virtual_int_registers_t bcm59040_vint_table_hibint[] = {
	{ BCM59040_REG_VINT1, BCM59040_REG_INT1M, BCM59040_INT1_HBINT }
};

virtual_int_registers_t bcm59040_vint_table_pmutoowarm[] = {
	{ BCM59040_REG_VINT1, BCM59040_REG_INT1M, BCM59040_INT1_PMUTOOWARM }
};

virtual_int_registers_t bcm59040_vint_table_chargerstatus[] = {
	{ BCM59040_REG_VINT2, BCM59040_REG_INT2M, (BCM59040_INT2_CHGINS |
											   BCM59040_INT2_CHGRM |
											   BCM59040_INT2_CHGOVERV |
											   BCM59040_INT2_USBINS |
											   BCM59040_INT2_USBRM |
											   BCM59040_INT2_USBOVERV ) },
	{ BCM59040_REG_VINT3, BCM59040_REG_INT3M, (BCM59040_INT3_VSROVERV |
											   BCM59040_INT3_VSROVERI |
											   BCM59040_INT3_VCHGRNOTOK |
											   BCM59040_INT3_CHG_WDT_ALARM |
											   BCM59040_INT3_VBUSLOWBND |
											   BCM59040_INT3_CHGWDTEXP |
											   BCM59040_INT3_IDOVERI) },
/*	{ BCM59040_REG_VINT7, BCM59040_REG_INT7M, (BCM59040_INT7_VBUS_VALID_F |
											   BCM59040_INT7_A_SESSVALID_F |
											   BCM59040_INT7_B_SESSEND_F |
											   BCM59040_INT7_ID_INSRT |
											   BCM59040_INT7_VBUS_VALID_R |
											   BCM59040_INT7_A_SESSVALID_R |
											   BCM59040_INT7_B_SESSEND_R |
											   BCM59040_INT7_ID_RMV) },
*/
	{ BCM59040_REG_VINT9, BCM59040_REG_INT9M, (BCM59040_INT9_RESUME_VBUS |
											   BCM59040_INT9_ID_CHNG) }
};

virtual_int_registers_t bcm59040_vint_table_regulatorstatus[] = {
	{ BCM59040_REG_VINT4, BCM59040_REG_INT4M, (BCM59040_INT4_LDO1OVRI |
											   BCM59040_INT4_LDO2OVRI |
											   BCM59040_INT4_LDO3OVRI |
											   BCM59040_INT4_LDO4OVRI |
											   BCM59040_INT4_LDO5OVRI |
											   BCM59040_INT4_LDO6OVRI) },
	{ BCM59040_REG_VINT5, BCM59040_REG_INT5M, (BCM59040_INT5_IOSROVRI |
											   BCM59040_INT5_CSROVRI |
											   BCM59040_INT5_IOSROVRV |
 											   BCM59040_INT5_CSROVRV) }
};

virtual_int_registers_t bcm59040_vint_table_battstatus[] = {
	{ BCM59040_REG_VINT2, BCM59040_REG_INT2M, BCM59040_INT2_EOC },
	{ BCM59040_REG_VINT4, BCM59040_REG_INT4M, (BCM59040_INT4_BBLOW |
											   BCM59040_INT4_FGC) },
	{ BCM59040_REG_VINT6, BCM59040_REG_INT6M, (BCM59040_INT6_MBTEMPFAULT |
											   BCM59040_INT6_MBTEMPLOW |
											   BCM59040_INT6_MBTEMPHIGH |
											   BCM59040_INT6_MBRM |
											   BCM59040_INT6_MBOV |
											   BCM59040_INT6_LOWBAT |
 											 BCM59040_INT6_VERYLOWBAT) },
	{ BCM59040_REG_VINT3, BCM59040_REG_INT3M, (BCM59040_INT3_CHGERRDIS | BCM59040_INT3_VBUSLOWBND) }
};

virtual_int_registers_t bcm59040_vint_table_rtc[] = {
	{ BCM59040_REG_VINT5, BCM59040_REG_INT5M, (BCM59040_INT5_RTCADJ |
											   BCM59040_INT5_RTC1S |
											   BCM59040_INT5_RTC60S |
 											   BCM59040_INT5_RTCA1) }
};

virtual_int_registers_t bcm59040_vint_table_usb[] = {
	{ BCM59040_REG_VINT2, BCM59040_REG_INT2M, BCM59040_INT2_CHGDET },
	{ BCM59040_REG_VINT7, BCM59040_REG_INT7M, (BCM59040_INT7_VBUS_VALID_F |
	/* Very important for Windows to recognize the device the first time... */	   BCM59040_INT7_B_SESSEND_F |
	/* For Poznan */								   BCM59040_INT7_ID_INSRT |
	/* Not sure why */								   BCM59040_INT7_VBUS_VALID_R |
	/* For device mode */								   BCM59040_INT7_A_SESSVALID_R |
	/* For Poznan */								   BCM59040_INT7_ID_RMV) },
/*	{ BCM59040_REG_VINT9, BCM59040_REG_INT9M, (BCM59040_INT9_RESUME_VBUS |
											   BCM59040_INT9_ID_CHNG) }
*/
};

virtual_int_registers_t bcm59040_vint_table_adc[] = {
	{ BCM59040_REG_VINT8, BCM59040_REG_INT8M, (BCM59040_INT8_SARCONVRDY0 |
											   BCM59040_INT8_SARCONVRDY1 |
											   BCM59040_INT8_SARCONVRDY2 |
											   BCM59040_INT8_SARCONVRDY3 |
											   BCM59040_INT8_SARCONVRDY4 |
											   BCM59040_INT8_SARCONVRDY5 |
											   BCM59040_INT8_SARCONVRDY6 |
											   BCM59040_INT8_SARCONVRDY7) },
	{ BCM59040_REG_VINT9, BCM59040_REG_INT9M, (BCM59040_INT9_SARCONVRDY8 |
											   BCM59040_INT9_SARCONVRDY9 |
											   BCM59040_INT9_SARCONVEND |
											   BCM59040_INT9_SARCONTCONVFAIL |
											   BCM59040_INT9_SARASYNCONVOFF |
											   BCM59040_INT9_SARASYNREQFAIL) }
};

/*
 * For speed in handling virtual interrupts this
 * table is expected to be in order of increasing
 * IRQ number! Do not break this rule or code will fail!
 */
virtual_int_entry_t bcm59040_vint_table[] = {
	{ BCM59040_IRQ_PONKEY,
	  ARRAY_SIZE(bcm59040_vint_table_ponkey),
	  bcm59040_vint_table_ponkey, 0 },

	{ BCM59040_IRQ_INVALID_HIB_MODE,
	  ARRAY_SIZE(bcm59040_vint_table_hibint),
	  bcm59040_vint_table_hibint, 0 },

	{ BCM59040_IRQ_PMU_TOO_WARM,
	  ARRAY_SIZE(bcm59040_vint_table_pmutoowarm),
	  bcm59040_vint_table_pmutoowarm, 0 },

	{ BCM59040_IRQ_CHARGER,
	  ARRAY_SIZE(bcm59040_vint_table_chargerstatus),
	  bcm59040_vint_table_chargerstatus, 0 },

	{ BCM59040_IRQ_REGULATORS,
	  ARRAY_SIZE(bcm59040_vint_table_regulatorstatus),
	  bcm59040_vint_table_regulatorstatus, 0 },

	{ BCM59040_IRQ_BATT,
	  ARRAY_SIZE(bcm59040_vint_table_battstatus),
	  bcm59040_vint_table_battstatus, 0 },

	{ BCM59040_IRQ_RTC,
	  ARRAY_SIZE(bcm59040_vint_table_rtc),
	  bcm59040_vint_table_rtc, 0 },

	{ BCM59040_IRQ_USB,
	  ARRAY_SIZE(bcm59040_vint_table_usb),
	  bcm59040_vint_table_usb, 0 },

	{ BCM59040_IRQ_ADC,
	  ARRAY_SIZE(bcm59040_vint_table_adc),
	  bcm59040_vint_table_adc, 0 }
};

struct pmu_client	*bcm59040_client_list[BCMPMU_MAX_CLIENT_TYPES] = {
	0
};

/*
 * Function declarations.
 */

static inline unsigned char bcm59040_reg_addr(int addr)
{
	if (addr >= BCM59040_PAGE1_BASE_ADDR)
		return (unsigned char)(addr - BCM59040_PAGE1_BASE_ADDR);
	else
		return (unsigned char)addr;
}

static int bcm59040_read_page(struct i2c_client *client, unsigned char *page)
{
	unsigned char 		reg_pagesel	= BCM59040_REG_PAGESEL;
	struct i2c_msg		msgs[BCM59040_I2C_MSG_LEN];
	int					nb_msg=0;
	int					ret=0, retries=3;

	msgs[BCM59040_I2C_ADDR_IND].addr	= client->addr;
	msgs[BCM59040_I2C_ADDR_IND].flags	= 0;
	msgs[BCM59040_I2C_ADDR_IND].len		= sizeof(reg_pagesel);
	msgs[BCM59040_I2C_ADDR_IND].buf		= &reg_pagesel;

	msgs[BCM59040_I2C_DATA_IND].addr	= client->addr;
	msgs[BCM59040_I2C_DATA_IND].flags	= I2C_M_RD;
	msgs[BCM59040_I2C_DATA_IND].len		= sizeof(*page);
	msgs[BCM59040_I2C_DATA_IND].buf		= page;

	while(retries--) {
		if((nb_msg = i2c_transfer(client->adapter,
								  msgs,
								  BCM59040_I2C_MSG_LEN)) == BCM59040_I2C_MSG_LEN)
			break;
	}

	if (nb_msg != BCM59040_I2C_MSG_LEN) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "Can't read the pagesel register!\n"));
		ret = -EIO;
	} else {
		PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX 
				  "read the pagesel register sucessfully [page = %d]\n", *page));
	}

	return ret;
}

static int bcm59040_write_page(struct i2c_client *client, unsigned char page)
{
	unsigned char 		reg_pagesel	= BCM59040_REG_PAGESEL;
	struct i2c_msg		msgs[BCM59040_I2C_MSG_LEN];
	int					nb_msg=0;
	int					ret=0, retries=3;

	msgs[BCM59040_I2C_ADDR_IND].addr	= client->addr;
	msgs[BCM59040_I2C_ADDR_IND].flags	= 0;
	msgs[BCM59040_I2C_ADDR_IND].len		= sizeof(reg_pagesel);
	msgs[BCM59040_I2C_ADDR_IND].buf		= &reg_pagesel;

	msgs[BCM59040_I2C_DATA_IND].addr	= client->addr;
	msgs[BCM59040_I2C_DATA_IND].flags	= I2C_M_NOSTART;
	msgs[BCM59040_I2C_DATA_IND].len		= sizeof(page);
	msgs[BCM59040_I2C_DATA_IND].buf		= &page;

	while(retries--) {
		if((nb_msg = i2c_transfer(client->adapter,
								  msgs,
								  BCM59040_I2C_MSG_LEN)) == BCM59040_I2C_MSG_LEN)
			break;
	}

	if (nb_msg != BCM59040_I2C_MSG_LEN) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "Can't write the pagesel register! [page = %d]\n", page));
		ret = -EIO;
	} else {
		PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX 
				"wrote the pagesel register sucessfully [page = %d]\n", page));
	}

	return ret;
}

static int bcm59040_select_page(bcm59040_ctxt_t *ctxt, int address)
{
	struct i2c_client	*client	= NULL;
	unsigned char		page	= EPAGE_0;
	int 				ret		= 0;

	BUG_ON(!ctxt);

	client = ctxt->client;
	BUG_ON(!client);

	if (address > BCM59040_PAGE_MAX_ADDR) {
		page = EPAGE_1;
	} else {
		page = EPAGE_0;
	}

	if (ctxt->page != page) {
		/* We have to set the appropriate page. */

		unsigned char actual_page = EPAGE_0;

		if (0 == (ret = bcm59040_read_page(client, &actual_page))) {
			switch (page)
			{
			  case EPAGE_0:
				actual_page &= 0xfe;
				break;
			  case EPAGE_1:
				actual_page |= 0x01;
				break;
			  default:
				PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "invalid page [page = %d]!\n", page));
				return -EINVAL;				
				break;
			}

			if (0 == (ret = bcm59040_write_page(client, actual_page))) {
				ctxt->page = page;
			}
		}
	}

	return ret;
}

static int bcm59040_read_data(bcm59040_ctxt_t *ctxt, int address, void *data, int len)
{
	struct i2c_msg		msgs[BCM59040_I2C_MSG_LEN];
	struct i2c_client	*client		= NULL;
	int					nb_msg = 0;
	unsigned char		reg_addr	= bcm59040_reg_addr(address);
	int					ret	= 0, retries = 3;

	BUG_ON(!ctxt);
	BUG_ON(!data);

	client = ctxt->client;
	BUG_ON(!client);

	msgs[BCM59040_I2C_ADDR_IND].addr	= client->addr;
	msgs[BCM59040_I2C_ADDR_IND].flags	= 0;
	msgs[BCM59040_I2C_ADDR_IND].len		= sizeof(reg_addr);
	msgs[BCM59040_I2C_ADDR_IND].buf		= &reg_addr;

	msgs[BCM59040_I2C_DATA_IND].addr	= client->addr;
	msgs[BCM59040_I2C_DATA_IND].flags	= I2C_M_RD;
	msgs[BCM59040_I2C_DATA_IND].len		= len;
	msgs[BCM59040_I2C_DATA_IND].buf		= data;

	while(retries--) {
		if((nb_msg = i2c_transfer(client->adapter,
								  msgs,
								  BCM59040_I2C_MSG_LEN)) == BCM59040_I2C_MSG_LEN)
			break;
	}

	if (nb_msg != BCM59040_I2C_MSG_LEN) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "i2c read failure!, %d\n", nb_msg));
		ret = -EIO;
	}

	return ret;
}

static int bcm59040_write_data(bcm59040_ctxt_t *ctxt, int address, void *data, int len)
{
	struct i2c_msg		msgs[BCM59040_I2C_MSG_LEN];
	struct i2c_client	*client		= NULL;
	int					nb_msg = 0;
	unsigned char		reg_addr	= bcm59040_reg_addr(address);
	int					ret	= 0, retries = 3;

	BUG_ON(!ctxt);
	BUG_ON(!data);

	client = ctxt->client;
	BUG_ON(!client);

	msgs[BCM59040_I2C_ADDR_IND].addr	= client->addr;
	msgs[BCM59040_I2C_ADDR_IND].flags	= 0;
	msgs[BCM59040_I2C_ADDR_IND].len		= sizeof(reg_addr);
	msgs[BCM59040_I2C_ADDR_IND].buf		= &reg_addr;

	msgs[BCM59040_I2C_DATA_IND].addr	= client->addr;
	msgs[BCM59040_I2C_DATA_IND].flags	= I2C_M_NOSTART;
	msgs[BCM59040_I2C_DATA_IND].len		= len;
	msgs[BCM59040_I2C_DATA_IND].buf		= data;

	while(retries--) {
		if((nb_msg = i2c_transfer(client->adapter,
								  msgs,
								  BCM59040_I2C_MSG_LEN)) == BCM59040_I2C_MSG_LEN)
			break;
	}

	if (nb_msg != BCM59040_I2C_MSG_LEN) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "i2c write failure! %d\n", nb_msg));
		ret = -EIO;
	}

	return ret;
}

/** bcm59040_bus_read - BCM59040 bus driver read
 *
 * This function reads data from the BCM59040 device over the I2C bus
 * connection. It is called via the bcm59040_bus_ops read callback. The
 * caller is the pmu device framework. Since BCM59040 has byte sized
 * registers any alignment and transfer size granularity is valid. We
 * never report an error on alignment. We should report an attempt to
 * read an invalid register.
 *
 * @adap:    Adapter to read data from.
 * @address: Register address within adapter I/O space.
 * @data:    Pointer to data buffer to hold read data.
 * @len:     Length of data to read in bytes.
 *
 * Returns 0 if successful otherwise standard Linux error codes
 *
 */

static int bcm59040_bus_read(struct pmu_adapter *adap, int address, void *data, int len)
{
	struct i2c_client	*client	= NULL;
	bcm59040_ctxt_t		*ctxt	= NULL;
	int			 ret 	= len;

	BUG_ON(!adap);
	BUG_ON(!data);

	client = container_of(adap->dev.parent, struct i2c_client, dev);
	BUG_ON(!client);

	ctxt = i2c_get_clientdata(client);
	BUG_ON(!ctxt);

	if (len > 10) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "can't read %d registers per burst!"
							"[max = 10]\n", len));
		return -EINVAL;	
	}

	if (in_atomic() || irqs_disabled()) {
		ret = mutex_trylock (&ctxt->mutex_i2c_rw);
		if (!ret) {
			printk ("bus_read: Bus activity is ongoing ... retry later: (%d|%d)\n", in_atomic(), irqs_disabled());
			return -EAGAIN;
		}
	} else {
		mutex_lock_nested (&ctxt->mutex_i2c_rw, SINGLE_DEPTH_NESTING);
	}

	if ((address & 0xFF00) == 0x0300) {
		if((address >= BCM59040_REG_VINT1) &&
		   (address <= BCM59040_REG_VINT9)) {
			int i,j;

			if(((address+len)-1) > BCM59040_REG_VINT9)
				len = (BCM59040_REG_VINT9 - address) + 1;
			ret = len;
			for(i=(address-BCM59040_REG_VINT1), j=0; j < len; i++,j++) {
				 *(((u8*)data)+j) = ctxt->int_regs[i];
			}
		} else if((address >= BCM59040_REG_VINT1M) &&
				  (address <= BCM59040_REG_VINT9M)) {
			int i,j;

			if(((address+len)-1) > BCM59040_REG_VINT9M)
				len = (BCM59040_REG_VINT9M - address) + 1;
			ret = len;
			for(i=(address-BCM59040_REG_VINT1M), j=0; j < len; i++,j++) {
				 *(((u8*)data)+j) = ctxt->masks[i];
			}
		} else {
			ret = -EINVAL;
		}
	} else if(0 == (ret = bcm59040_select_page(ctxt, address))) {
		if((ret = bcm59040_read_data(ctxt, address, data, len) == 0)) {
			/* Read operation successful, return the number of byte read */
			ret = len;
		}
	}

	mutex_unlock (&ctxt->mutex_i2c_rw);

	return ret;
}

/** bcm59040_bus_write - BCM59040 bus driver write
 *
 * This function writes data to the BCM59040 device over the I2C bus
 * connection. It is called via the bcm59040_bus_ops write callback. The
 * caller is the pmu device framework. Since BCM59040 has byte sized
 * registers any alignment and transfer size granularity is valid. We
 * never report an error on alignment. We should report an attempt to
 * write an invalid register.
 *
 * @adap:    Adapter to write data to.
 * @address: Register address within adapter I/O space.
 * @data:    Pointer to data buffer holding write data.
 * @len:     Length of data to write in bytes.
 *
 * Returns 0 if successful otherwise standard Linux error codes
 *
 */
static int bcm59040_bus_write(struct pmu_adapter *adap, int address, void *data, int len)
{
	struct i2c_client	*client	= NULL;
	bcm59040_ctxt_t		*ctxt	= NULL;
	int					ret		= 0;

	BUG_ON(!adap);
	BUG_ON(!data);

	client = container_of(adap->dev.parent, struct i2c_client, dev);
	BUG_ON(!client);

	ctxt = i2c_get_clientdata(client);
	BUG_ON(!ctxt);

	if (len > 10) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "can't write %d registers per burst!"
							"[max = 10]\n", len));
		return -EINVAL;	
	}

	if (in_atomic() || irqs_disabled()) {
		ret = mutex_trylock (&ctxt->mutex_i2c_rw);
		if (!ret) {
			printk ("bus_write: Bus activity is ongoing ... retry later (%d|%d)\n", in_atomic(), irqs_disabled());
			return -EAGAIN;
		}
	} else {
		mutex_lock_nested (&ctxt->mutex_i2c_rw, SINGLE_DEPTH_NESTING);
	}

	if((address & 0xFF00) == 0x0300) {
		if((address >= BCM59040_REG_VINT1) &&
		   (address <= BCM59040_REG_VINT9)) {
			int i,j;

			if(((address+len)-1) > BCM59040_REG_VINT9)
				len = (BCM59040_REG_VINT9 - address) + 1;
			ret = len;
			for(i=(address-BCM59040_REG_VINT1), j=0; j < len; i++,j++) {
				 ctxt->int_regs[i] = *(((u8*)data)+j);
			}
		} else if((address >= BCM59040_REG_VINT1M) &&
				  (address <= BCM59040_REG_VINT9M)) {
			int i,j;

			if(((address+len)-1) > BCM59040_REG_VINT9M)
				len = (BCM59040_REG_VINT9M - address) + 1;
			ret = len;
			for(i=(address-BCM59040_REG_VINT1M), j=0; j < len; i++,j++) {
				 ctxt->masks[i] = *(((u8*)data)+j);
			}
			if (0 == (ret = bcm59040_select_page(ctxt,
						BCM59040_REG_INT1M+(address-BCM59040_REG_VINT1M)))) {
				if (0 == (ret = bcm59040_write_data(ctxt,
						BCM59040_REG_INT1M+(address-BCM59040_REG_VINT1M),
						data, len))) {
					/* Write operation successful, return the number of byte written */
					ret = len;	
				}
			}
		} else {
			ret = -EINVAL;
		}
	} else if (0 == (ret = bcm59040_select_page(ctxt, address))) {
		if (0 == (ret = bcm59040_write_data(ctxt, address, data, len))) {
			/* Write operation successful, return the number of byte written */
			ret = len;	
		}
	}

	mutex_unlock(&ctxt->mutex_i2c_rw);

	return ret;
}

static const struct pmu_algorithm bcm59040_algorithm = {
	.read	= bcm59040_bus_read,
	.write	= bcm59040_bus_write,
};

static int bcm59040_set_irq_mask(unsigned int irq, e_irq_mask mask)
{
	bcm59040_ctxt_t						*ctxt	= get_irq_chip_data(irq);
	struct i2c_client					*client	= NULL;
	struct pmu_bcm59040_platform_data	*pdata	= NULL;
	int									nb_byte = 0, i;
	u8									reg_int = 0,reg_val, reg_msk = 0;

	BUG_ON(!ctxt);

	client = ctxt->client;
	BUG_ON(!client);

	pdata = client->dev.platform_data;
	BUG_ON(!pdata);

	irq -= pdata->virqs.start;
	for(i=0; i<bcm59040_vint_table[irq].num_registers ; i++) {
		reg_int = bcm59040_vint_table[irq].reg_list[i].mask_reg_addr;
		if((nb_byte = bcm59040_bus_read(&ctxt->adapter,
									 reg_int,
									 &reg_val,
									 sizeof(reg_val))) < 0) {
			PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "error reading mask register [irq=%d/%d] !\n", irq, i));
			return -EIO;
		}
		reg_msk = bcm59040_vint_table[irq].reg_list[i].reg_mask;
		if(mask == EIRQ_MASK) {
			reg_val |= reg_msk;	/* mask all related bits for this IRQ */
		} else {
			reg_val &= ~reg_msk;
		}

		if((nb_byte = bcm59040_bus_write(&ctxt->adapter,
									 reg_int,
									 &reg_val,
									 sizeof(reg_val))) < 0) {
			PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "mask setting failure [irq=%d/%d] !\n", irq, i));
			return -EIO;
		}
	}

	if(mask == EIRQ_MASK)
		ctxt->masked[irq] = 1;
	else
		ctxt->masked[irq] = 0;

	return 0;
}

/** bcm59040_irq_ack - IRQChip function to acknowledge an interrupt
 *
 * For BCM59040 these are actually "virtual" interrupts we're being
 * requested to disable. Of course virtual interrupts need to be
 * translated into the appropriate physical interrupts here so we
 * can acknowledge the proper the hardware interrupts.
 *
 * @irq:    number of BCM59040 virtual interrupt to ack
 *
 * @note Since several BCM59040 interrupts may be hidden beneath
 *       a single virtual interrupt we may actually acknowledge several physical
 *       interrupts when we acknowledge the specified virtual interrupts.
 *
 * @todo	Needs to be implemented.
 *
 */
static void bcm59040_irq_ack(unsigned irq)
{
	// Since this function is ONLY called with interrupts disabled, we don't
	// need to disable irqs around the following
	return;
}

/** bcm59040_irq_mask - IRQChip function to mask, i.e., disable an interrupt
 *
 * For BCM59040 these are actually "virtual" interrupts we're being
 * requested to disable. Of course virtual interrupts need to be
 * translated into the appropriate physical interrupts here so we
 * can configure the hardware appropriately.
 *
 * @irq:	Number of BCM59040 virtual interrupt to mask.
 *
 * @note Since several BCM59040 interrupts may be hidden beneath
 *       a single virtual interrupt we may actually disable several physical
 *       interrupts when we turn off the specified virtual interrupts.
 *
 */
static void bcm59040_irq_mask(unsigned irq)
{
	bcm59040_set_irq_mask(irq, EIRQ_MASK);
}

/** bcm59040_irq_unmask - IRQChip function to unmask, i.e., enable an interrupt
 *
 * For BCM59040 these are actually "virtual" interrupts we're being
 * requested to enable. Of course virtual interrupts need to be
 * translated into the appropriate physical interrupts here so we
 * can configure the hardware appropriately.
 *
 * @irq:		Number of BCM59040 virtual interrupt to unmask.
 *
 * @note Since several BCM59040 interrupts may be hidden beneath
 *       a single virtual interrupt we may actually enable several physical
 *       interrupts when we turn off the specified virtual interrupts.
 *
 */
static void bcm59040_irq_unmask(unsigned irq)
{
	bcm59040_set_irq_mask(irq, EIRQ_UNMASK);
}

/** bcm59040_irq_set_wake - IRQChip function to set a BCM59040 interrupt as a wakeup event
 *
 * For BCM59040 these are actually "virtual" interrupts we're being
 * requested to disable. Of course virtual interrupts need to be
 * translated into the appropriate physical interrupts here so we
 * can configure the hardware appropriately. Since we are configuring
 * what interrupts are wakeup events this means we need to mask all
 * interrupts except for these when we enter suspend. Conversely we
 * have to remember what the mask/unmask status was when we entered
 * suspend and restore it on resume.
 *
 * @irq:		Number of BCM59040 virtual interrupt to set as wakeup event.
 * @on			0 to disable irq as wakeup event, 1 to enable as wakeup event.
 *
 * Return 0 if no error.
 *
 * @note    Since several BCM59040 interrupts may be hidden beneath
 *          a single virtual interrupt we may actually choose several physical
 *          interrupts as wakeup events as a response for configuring the
 *          specified virtual interrupt as a wakeup event.
 *
 * @note    The way this ought to work is that others should enable or disable a
 *          BCM59040 IRQ event as a wakeup event. This function should track the
 *          number of interrupts selected as wakeup events. When the count total
 *          becomes zero (and initially) the primary wakeup event into the SoC
 *          should be disabled also. When the count transitions from zero to one
 *          or greater the primary wakeup event into the SoC should be enabled. In
 *          other words hide the need to manage the wakeup event into the SoC.
 *
 * @todo	Needs to be implemented.
 *
 */
static int bcm59040_irq_set_wake(unsigned irq, unsigned on)
{
	bcm59040_ctxt_t				*ctxt=get_irq_chip_data(irq);
	struct i2c_client			*client=NULL;
	struct pmu_bcm59040_platform_data	*pdata=NULL;

	BUG_ON(!ctxt);
	client = ctxt->client;
	BUG_ON(!client);
	pdata = client->dev.platform_data;
	BUG_ON(!pdata);
	irq -= pdata->virqs.start;
	bcm59040_vint_table[irq].wakeup=(on ? 1 : 0);
	return 0;
} // bcm59040_irq_set_wake


static struct irq_chip bcm59040_irq_chip =
{
	.ack		= bcm59040_irq_ack,
	.mask		= bcm59040_irq_mask,
	.unmask		= bcm59040_irq_unmask,
	.set_wake	= bcm59040_irq_set_wake,
};

static irqreturn_t pmu_bus_handle_irq(int irq, void *data)
{
	bcm59040_ctxt_t *ctxt = (bcm59040_ctxt_t *)data;

	disable_irq(irq);

	up(&ctxt->sem_isr);

	return IRQ_HANDLED;
}

#if 0
static irqreturn_t pmu_bus_handle_virtual_irq(int irq, void *data)
{
	/* virtual IRQs from the PMU: 
	   - for now, each of the 72 bits have their own handler.
	   - the grouping of the interrupt is not yet implemented. */

	bcm59040_ctxt_t *ctxt = (bcm59040_ctxt_t *)data;
	PMU_DEBUG(DBG_TRACE, ("Handler virtual IRQ from dev [%s]\n", ctxt->client->name));
	
	return IRQ_HANDLED;
}
#endif

static int bcm59040_pmu_irqs_apply_masks(bcm59040_ctxt_t * ctxt)
{
	int nb_byte;

	BUG_ON(!ctxt);

	nb_byte = bcm59040_bus_write(&ctxt->adapter, BCM59040_REG_INT1M,
								 ctxt->masks, BCM59040_NUM_INT_REG);

	if (nb_byte != BCM59040_NUM_INT_REG) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "virqs mask initialization failure!\n"));
		return -EIO;
	}
	
	PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "virqs masks applied\n"));

	return 0;
}


static int bcm59040_pmu_irqs_read_masks(bcm59040_ctxt_t * ctxt)
{
	int nb_byte;
	int i;
	unsigned char read_masks[BCM59040_NUM_INT_REG];

	BUG_ON(!ctxt);

	nb_byte = bcm59040_bus_read(&ctxt->adapter, BCM59040_REG_INT1M, 
								read_masks, BCM59040_NUM_INT_REG);

	if (nb_byte != BCM59040_NUM_INT_REG) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "virqs read failure!\n"));
		return -EIO;
	}
	
	PMU_DEBUG(DBG_ERROR, ("read_masks: "));
	for(i = 0 ; i < BCM59040_NUM_INT_REG ; i ++) {
		PMU_DEBUG(DBG_ERROR, ("%d, ", read_masks[i]));
	}
	PMU_DEBUG(DBG_ERROR, ("\n"));

	return 0;
}

/** bcm59040_handle_virq - Process virtual BCM59040 PMU interrupts
 *
 * This function will scan the virtual (saved copy) of the interrupt
 * registers for the BCM59040 PMU and generate the different virtual
 * interrupts appropriate depending upon the interrupt status bits
 * that are set.
 *
 */
void bcm59040_handle_virq(bcm59040_ctxt_t *ctxt)
{
	int		int_vector,i,j;

	PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "Processing BCM59040 virtual interrupts:\n"));

	for(i=0 ; i<ARRAY_SIZE(bcm59040_vint_table) ; i++) {

		int_vector = 0;			/* zero means no int to be generated */

		for(j=0 ; j<bcm59040_vint_table[i].num_registers; j++) {

			uint8_t	intreg, mask, reg_mask;

			intreg=ctxt->int_regs[bcm59040_vint_table[i].reg_list[j].reg_addr-BCM59040_REG_VINT1];
			mask=~ctxt->masks[bcm59040_vint_table[i].reg_list[j].mask_reg_addr-BCM59040_REG_INT1M];
			reg_mask=bcm59040_vint_table[i].reg_list[j].reg_mask;

			PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "Checking register %02X having a value of %02X\n",
								  bcm59040_vint_table[i].reg_list[j].reg_addr,
								  ctxt->int_regs[bcm59040_vint_table[i].reg_list[j].reg_addr-BCM59040_REG_VINT1]));

			if((ctxt->int_regs[bcm59040_vint_table[i].reg_list[j].reg_addr-BCM59040_REG_VINT1] &
			    ~ctxt->masks[bcm59040_vint_table[i].reg_list[j].mask_reg_addr-BCM59040_REG_INT1M]) &
				bcm59040_vint_table[i].reg_list[j].reg_mask) {
				int_vector = bcm59040_vint_table[i].int_vector;

				if(int_vector) {
					PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "Dispatching virtual interrupt %d\n", int_vector));
					generic_handle_irq(int_vector);
				} else {
					PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "Skipping virtual interrupt %d\n",
									bcm59040_vint_table[i].int_vector));
				}
				break;
			}
		}

	}

	PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "Completed processing BCM59040 virtual interrupts:\n"));
}

static int ktask_fn(void *data)
{
	bcm59040_ctxt_t				*ctxt = NULL;
	struct i2c_client			*client = NULL;
	struct pmu_bcm59040_platform_data	*pdata = NULL;

	BUG_ON(!data);

	ctxt = data;

	client = ctxt->client;
	BUG_ON(!client);

	pdata = client->dev.platform_data;
	BUG_ON(!pdata);

	PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "kernel process spawned\n"));

	while(! kthread_should_stop()) {

		if (down_interruptible(&ctxt->sem_isr)) {
			printk(PFX "ktask received a signal\n");
			return -ERESTART;
		}

		do {
			int	retval[2]={0,0};

			for( ; ; )
			{
				if( retval[0] != sizeof( ctxt->int_regs ) )
					retval[0]=bcm59040_bus_read(&ctxt->adapter, BCM59040_REG_INT1,
								 ctxt->int_regs, sizeof(ctxt->int_regs));
				if( retval[1] != sizeof( ctxt->masks ) )
					retval[1]=bcm59040_bus_read(&ctxt->adapter, BCM59040_REG_INT1M,
								 ctxt->masks, sizeof(ctxt->masks));

				if( (retval[0] == sizeof( ctxt->int_regs )) && (retval[1] == sizeof( ctxt->masks )) )
					break;

				printk(KERN_ERR PFX "can't get interrupt status.. trying again!: %d, %d\n", retval[0], retval[1]);

				msleep( 50 );
			}
			bcm59040_handle_virq(ctxt);
			

		/* CLEANUP: gpio_get_value() should not be called here. temporary. */
		} while ((! gpio_get_value(1)) && (! kthread_should_stop()));

		enable_irq(client->irq);
	}

	return 0;
}

static bcm59040_ctxt_t * bcm59040_ctxt_setup(struct i2c_client *client)
{
	bcm59040_ctxt_t *ctxt = NULL;
	int i;

//	unsigned char masks[BCM59040_NUM_INT_REG] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}; doesn't work
//	unsigned char masks[BCM59040_NUM_INT_REG] = {0xFF, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}; works
//	unsigned char masks[BCM59040_NUM_INT_REG] = {0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}; works

	BUG_ON(!client);

	PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "Context initialization\n"));

	ctxt = (bcm59040_ctxt_t *) kzalloc(sizeof(bcm59040_ctxt_t), GFP_KERNEL);
	if(NULL == ctxt) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "Can't allocate work!\n"));
		return NULL;
	}

	mutex_init(&ctxt->mutex_i2c_rw);
	sema_init(&ctxt->sem_isr, 0);

	ctxt->client = client;
	i2c_set_clientdata(client, ctxt);

	/* virqs masks initialization: Initially, all virqs */
	/* are masked.                                      */
	for(i = 0 ; i < BCM59040_NUM_INT_REG ; i++) {
		ctxt->masks[i] = 0xFF;
	}

	PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "Kernel thread creation\n"));

	ctxt->ktask = kthread_create(ktask_fn, ctxt, client->name);
	if (IS_ERR(ctxt->ktask)) {
		int ret = PTR_ERR(ctxt->ktask);
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "Kernel thread creation failure! err=%d\n", ret));
		kfree(ctxt);
		return NULL;
	}

	/* Preparing for Bus Backend Registration. */
	ctxt->adapter.owner	= THIS_MODULE;
	ctxt->adapter.algo	= &bcm59040_algorithm;

	/* set up the sysfs linkage to our parent device */
	ctxt->adapter.dev.parent = &client->dev;
	strncpy(ctxt->adapter.name, client->name, 
			sizeof(ctxt->adapter.name));

	return ctxt;
}

static void bcm59040_ctxt_cleanup(bcm59040_ctxt_t * ctxt)
{
	kthread_stop(ctxt->ktask);
	
	kfree(ctxt);
}

static int bcm59040_pmu_irqs_setup(bcm59040_ctxt_t * ctxt)
{
	struct pmu_bcm59040_platform_data	*pdata = NULL;
	int								i;

	BUG_ON(!ctxt);
	BUG_ON(!ctxt->client);

	pdata = ctxt->client->dev.platform_data;
	BUG_ON(!pdata); 

	PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "setup virqs\n"));

	for (i = 0; i <= (pdata->virqs.end - pdata->virqs.start) ; i++) {
		unsigned int irq = pdata->virqs.start + i;

		dynamic_irq_init(irq);

		set_irq_chip(irq, &bcm59040_irq_chip);
		set_irq_chip_data(irq, ctxt);

		set_irq_handler(irq, handle_level_irq);
		set_irq_flags(irq, IRQF_VALID | IRQF_PROBE);

	}

	bcm59040_pmu_irqs_read_masks(ctxt);

	return 0;
}

static int bcm59040_pmu_irqs_cleanup(bcm59040_ctxt_t * ctxt)
{
	struct pmu_bcm59040_platform_data	*pdata = NULL;

	int	i;

	BUG_ON(!ctxt);
	BUG_ON(!ctxt->client);

	pdata = ctxt->client->dev.platform_data;
	BUG_ON(!pdata); 

	PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "setup virqs\n"));

	for (i = 0; i < (pdata->virqs.end - pdata->virqs.start) ; i++) {
		dynamic_irq_cleanup(pdata->virqs.start + i);
	}

	for (i = 0; i < (pdata->virqs.end - pdata->virqs.start)  ; i++) {
		free_irq(pdata->virqs.start + i, ctxt);
	}

	return 0;
}

/** bcm_get_pclient_from_client_type - Obtain PMU_CLIENT from a supplied type.
 *
 * This function helps isolate different PMU client drivers from each other. Rather
 * than hardcoding variables and such each client can call this function to obtain
 * the pmu_client value of one of it's sibling devices/drivers.
 *
 * @pclient:		struct pmu_client of caller (for possible verification)
 * @client_type:	type ID of pmu_client to look up.
 *
 * Returns pclient pointer or NULL if no valid matching entry found.
 *
 */
struct pmu_client *bcm_get_pclient_from_client_type(struct pmu_client *pclient,
													bcmpmu_clients_t client_type)
{
	int		i;

	if(!pclient || (client_type > BCMPMU_MAX_CLIENT_TYPES))
		return NULL;

	for(i=0; i<BCMPMU_MAX_CLIENT_TYPES; i++) {
		if(pclient == bcm59040_client_list[i])
			break;
	}

	if(i == BCMPMU_MAX_CLIENT_TYPES) {
		return NULL;
	} else {
		PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "type=%d, pclient=%08X\n",
						client_type,
						(uint32_t)bcm59040_client_list[client_type]));
		return bcm59040_client_list[client_type];
	}
}
EXPORT_SYMBOL_GPL(bcm_get_pclient_from_client_type);

/** bcm59040_i2c_probe - probe for BCM59040 I2C client
 *
 * This function checks for existence of a BCM59040 PMU and initializes
 * the I2C connection and driver routines if one is found. It also adds
 * the PMU client devices to the PMU framework so the PMU drivers can
 * register and load against them.
 */
static int bcm59040_i2c_probe(struct i2c_client *client, const struct i2c_device_id * id)
{
	bcm59040_ctxt_t *ctxt = NULL;
	struct pmu_bcm59040_platform_data *pdata =
		(struct pmu_bcm59040_platform_data*)client->dev.platform_data;
	int ret = 0;

	BUG_ON(!client);

	ctxt = bcm59040_ctxt_setup(client);

	{
		u8	pmu_id;

		ret = bcm59040_bus_read(&ctxt->adapter,
								BCM59040_PAGE1_BASE_ADDR+BCM59040_REG_PMUID,
								&pmu_id, sizeof(pmu_id));
		if(ret != sizeof(pmu_id)) {
			PMU_DEBUG(DBG_ERROR, ("Failed to read BCM59040 chip ID!\n"));
			bcm59040_ctxt_cleanup(ctxt);
			return -EINVAL;
		}
		PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "BCM59040 chip ID: %02X\n", (unsigned)pmu_id));

		if ((pmu_id != PMU_59040_B0_REV) &&
			(pmu_id != PMU_59040_B1_REV)) {
			PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "PMU Chip Version and software does not match\n"));
			bcm59040_ctxt_cleanup(ctxt);
			return -EINVAL;
		}
	}

	{
		int				i;
		unsigned char	int_reg[BCM59040_NUM_INT_REG];

		PMU_DEBUG(DBG_TRACE, ("Initial interrupt status reg: "));
		ret = bcm59040_bus_read(&ctxt->adapter, BCM59040_REG_INT1, int_reg, sizeof(int_reg));
		if(ret != sizeof(int_reg)) {
			PMU_DEBUG(DBG_ERROR, ("Failed to read BCM59040 INT1 register!\n"));
			bcm59040_ctxt_cleanup(ctxt);
			return -EINVAL;
		}

		for (i = 0; i < BCM59040_NUM_INT_REG; i++) {
			PMU_DEBUG(DBG_TRACE, (" %x", (int)int_reg[i]));
		}
		PMU_DEBUG(DBG_TRACE, ("\n"));
	}

	if (0 != (ret = bcm59040_pmu_irqs_apply_masks(ctxt))) {
		bcm59040_ctxt_cleanup(ctxt);
		return ret;
	}

	{
		uint8_t		otgctrl[2];

		/* Default set the OTG controller part off. If we need it, it is enabled by it's driver. */
		ret = bcm59040_bus_read( &ctxt->adapter, BCM59040_REG_OTGCTRL1, otgctrl, sizeof( otgctrl ) );
		if( ret != sizeof( otgctrl ) )
		{
			printk( KERN_WARNING PFX "Unable to read OTGCTRL registers!\n" );
			bcm59040_ctxt_cleanup(ctxt);
			return -EINVAL;
		}

		otgctrl[0]&=0x7F;
		otgctrl[1]&=0xBF;

		ret = bcm59040_bus_write( &ctxt->adapter, BCM59040_REG_OTGCTRL1, otgctrl, sizeof( otgctrl ) );
		if( ret != sizeof( otgctrl ) )
		{
			printk( KERN_WARNING PFX "Unable to initialize OTGCTRL registers!\n" );
			bcm59040_ctxt_cleanup(ctxt);
			return -EINVAL;
		}
	}

	{
		int	i;
		unsigned char	int_reg[BCM59040_NUM_INT_REG];

		PMU_DEBUG(DBG_ERROR, ("Initial interrupt status reg: "));
		bcm59040_bus_read(&ctxt->adapter, BCM59040_REG_INT1, int_reg, sizeof(int_reg));

		for (i = 0; i < BCM59040_NUM_INT_REG; i++) {
			PMU_DEBUG(DBG_ERROR, (" %x", (int)int_reg[i]));
		}
		PMU_DEBUG(DBG_ERROR, ("\n"));
	}


	bcm59040_pmu_irqs_setup(ctxt);

	PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "wake up process\n"));
	wake_up_process(ctxt->ktask);

	{
		if(0 != request_irq(client->irq, pmu_bus_handle_irq, IRQF_DISABLED | IRQF_TRIGGER_LOW, 
						client->name, ctxt)) {
			PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "Can't request irq \"%d\"\n", client->irq));
			bcm59040_pmu_irqs_cleanup(ctxt);
			bcm59040_ctxt_cleanup(ctxt);
			return -ENXIO;
		}

		enable_irq_wake(client->irq);
	}

	ret = pmu_add_adapter(&ctxt->adapter);
	if (0 != ret) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX "Can't register backend!\n"));
		free_irq(client->irq, ctxt);
		bcm59040_pmu_irqs_cleanup(ctxt);
		bcm59040_ctxt_cleanup(ctxt);
		return -ENXIO;
	}

	{
		int			i;
		struct pmu_client	*pclient;
		bcm59040_client_platform_data_t		*cpdata;

		for(i=0 ; i<pdata->pmu_board_info_entries; i++) {
			pclient = pmu_register_device(&pdata->pmu_board_info[i]);
			if(!pclient) {
				free_irq(client->irq, ctxt);
				bcm59040_pmu_irqs_cleanup(ctxt);
				(void)pmu_del_adapter(&ctxt->adapter);
				bcm59040_ctxt_cleanup(ctxt);
				return -ENXIO;
			}

			/*
			 * Register client into our tracking table for possible lookup later.
			 */

			cpdata = pdata->pmu_board_info[i].platform_data;
			if(cpdata->base.client_type < BCMPMU_MAX_CLIENT_TYPES) {
				bcm59040_client_list[cpdata->base.client_type] = pclient;
				PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "type=%d, pclient=%08X\n",
									  cpdata->base.client_type,
									  (uint32_t)pclient));
			}
		}
	}

	printk(KERN_INFO PFX "BCM59040 pmu bus backend driver, Copyright (c) 2010 TomTom B.V and Broadcom Corporation\n");
	return 0;
}

static int bcm59040_i2c_remove(struct i2c_client *client)
{
	bcm59040_ctxt_t *ctxt = i2c_get_clientdata(client);

	pmu_del_adapter(&ctxt->adapter);
	bcm59040_pmu_irqs_apply_masks(ctxt);
	free_irq(client->irq, ctxt);
	bcm59040_pmu_irqs_cleanup(ctxt);
	bcm59040_ctxt_cleanup(ctxt);

	PMU_DEBUG(DBG_TRACE, (KERN_INFO PFX "removed\n"));
	
	return 0;
}

#ifdef CONFIG_PM
static uint8_t	bcm59040_intmask_state[9]={0,0,0,0,0,0,0,0,0};
static int bcm59040_i2c_suspend(struct i2c_client *client, pm_message_t state)
{
	struct pmu_bcm59040_platform_data	*pdata=(struct pmu_bcm59040_platform_data*)client->dev.platform_data;
	bcm59040_ctxt_t				*ctxt=get_irq_chip_data(pdata->virqs.start);
	int					irq_idx;
	int ret;

	ret = 0;

	/* First store all INTMSK registers. We'll restore them exactly as this on resume. */
	if( bcm59040_read_data( ctxt, BCM59040_REG_INT1M, bcm59040_intmask_state, sizeof( bcm59040_intmask_state ) ) < 0 )
	{
		printk( KERN_ERR PFX "Error suspending PMU!\n" );
		return -EIO;
	}

	/* Now check our wakeupsources. If it is set as a wakeupsource, enable it otherwise disable it. */
	for( irq_idx=pdata->virqs.start; irq_idx <= pdata->virqs.end; irq_idx++ )
	{
		if( bcm59040_vint_table[irq_idx - pdata->virqs.start].wakeup )
			bcm59040_irq_unmask( irq_idx );
		else
			bcm59040_irq_mask( irq_idx );
	}

	/* If we support regulators, put it in the right state. */
	printk ("Suspend processed for bcm59040-core\n");
#ifdef CONFIG_REGULATOR
	if( state.event == PM_EVENT_SUSPEND )
		ret = regulator_suspend_prepare( PM_SUSPEND_MEM );
#endif

	return ret;
}

static int bcm59040_i2c_resume(struct i2c_client *client)
{
	struct pmu_bcm59040_platform_data	*pdata=(struct pmu_bcm59040_platform_data*)client->dev.platform_data;
	bcm59040_ctxt_t				*ctxt=get_irq_chip_data(pdata->virqs.start);

	/* Resume the INTMSK registers. */
	if (bcm59040_write_data( ctxt, BCM59040_REG_INT1M, bcm59040_intmask_state, sizeof( bcm59040_intmask_state ) ) < 0 )
	{
		printk( KERN_ERR PFX "Error resuming PMU!\n" );
		return -EIO;
	}

	printk ("Resume processed for bcm59040-core\n");

	return 0;
}

#else
#define bcm59040_i2c_suspend	NULL
#define bcm59040_i2c_resume	NULL
#endif /* CONFIG_PM	*/

static const struct i2c_device_id bcm59040_id[] = {
	{ BCM59040_DEVNAME, 1, },
	{ }
};

/* BCM59040 core driver registration data
 *
 */

static struct i2c_driver bcm59040_i2c_driver =
{
	.id		= /*I2C_DRIVERID_PMU*/ /*I2C_DRIVERID_BCM59040*/1049, /* CLEANUP ... */
	.probe	= bcm59040_i2c_probe,
	.remove	= bcm59040_i2c_remove,
	.suspend= bcm59040_i2c_suspend,
	.resume	= bcm59040_i2c_resume,
	.driver = {
		.name	= BCM59040_DEVNAME,
		.owner	= THIS_MODULE,
	},
	.id_table	= bcm59040_id,
};

/** bcm59040_core_init - BCM59040 core module Initialization
 *
 * Registers the BCM59040 core driver with the PMU device
 * framework.
 *
 * @return 0 if successful otherwise standard Linux error codes.
 *
 */

static int __init bcm59040_core_init(void)
{
	int err;

    if ((err = i2c_add_driver(&bcm59040_i2c_driver))) {
		PMU_DEBUG(DBG_ERROR, (KERN_ERR PFX
				  "Could Not Be Added. Err Code: [%i]\n", err));
		return err;
	}

	PMU_DEBUG(DBG_INFO, (KERN_INFO PFX "i2c driver registered\n"));

	return 0;
}

/** bcm59040_core_exit - BCM59040 core module Exit
 *
 * Deregisters the BCM59040 core driver with the PMU device
 * framework.
 *
 * @return 0 if successful otherwise standard Linux error codes.
 *
 */

static void __exit bcm59040_core_exit(void)
{
	i2c_del_driver(&bcm59040_i2c_driver);
}


subsys_initcall(bcm59040_core_init);
module_exit(bcm59040_core_exit);

MODULE_DESCRIPTION(BCM59040_CORE_MOD_DESCRIPTION);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Broadcom Corporation and TomTom B.V");
MODULE_VERSION(BCM59040_CORE_MOD_VERSION);
