/* psoc-ctsic.h
 *
 * Control driver for PSoc Capacitive Touch Screen IC
 *
 * Copyright (C) 2009 TomTom BV <http://www.tomtom.com/>
 * Authors: Vincent Dejouy <vincent.dejouy@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef INCLUDE_LINUX_PSOC_CTSIC_H
#define INCLUDE_LINUX_PSOC_CTSIC_H

#define HST_MODE_REGISTER_OFFSET                        0                     
#define HST_MODE_REGISTER_DEVICE_MODE_MASK              0x70                  
#define HST_MODE_REGISTER_DATA_RD_TOGGLE_MASK           0x80                  
#define HST_MODE_REGISTER_SOFT_RESET_MASK               0x01                  
#define TT_MODE_REGISTER_OFFSET				1
#define TT_MODE_REGISTER_BOOTLOADER_MODE_MASK		0x10
#define TT_MODE_REGISTER_BUFFER_INVALID_MASK		0x20
#define TT_STAT_REGISTER_TOUCHES_MASK                   0x0F                  
#define GEST_COUNT_REGISTER_GEST_COUNT_MASK             0x7F                  
#define HST_MODE_REGISTER_DEVICE_MODE_OFFSET            4                     
#define HST_MODE_REGISTER_DATA_RD_TOGGLE_OFFSET         7
#define TT_MODE_REGISTER_BUFFER_INVALID_OFFSET		5
#define OPERATING_MODE                                  0                     
#define SYS_INFO_MODE                                   1                     
#define RAW_COUNTS_TEST_MODE                            4                     
#define DIFFERENCE_COUNTS_TEST_MODE                     5                     
#define IDAC_SETTINGS_TEST_MODE                         6                     
#define BASELINE_TEST_MODE                              7                     
#define DONT_TOGGLE_DATA_RD_BIT                         0                     
#define TOGGLE_DATA_RD_BIT                              1                     
#define DELAY_BETWEEN_SET_OFFSET_REG_AND_RW_REG_VALUE   200 //in usecs        
#define DELAY_TO_START_TS_APPLICATION                   1000 //in msecs       
#define DELAY_TO_SWITCH_TO_SYS_INFO_MODE		1000 //in msecs
#define DELAY_TO_RUN_AUTO_RECALIBRATION			1000 //in msecs
#define VERBOSE_TUNING                                  3                     
#define VERBOSE_DEBUG                                   2                     
#define VERBOSE_ERROR                                   1                     
#define VERBOSE_DISABLED                                0                    
#define XMIN                                            0                     
#define XMAX                                            0xDC                  
#define YMIN                                            0                     
#define YMAX                                            0x140                 
#define PRESSUREMIN                                     0                     
#define PRESSUREMAX                                     1                     
#define GESTUREMIN                                      0                     
#define GESTUREMAX                                      0x10                  
#define FINGERSMIN                                      0                     
#define FINGERSMAX                                      2

/*Warning: Number of touches is a field in a register of the TS                 
  But it means:                                                                 
  0       No touch detected                                             
  1       One touch detected                                            
  2       Two touches detected with ghosting                            
  3       Two touches detected with no ghosting                         
  4 to 15 Reserved                                                      
*/                                                                            

#define MAXNUMTOUCHES                                   3                     
#define ROWS                                            11                    
#define COLUMNS                                         19                    

#define MAX_CHAR_PER_LINE       			200
#define MAX_BYTE_PER_LINE       			MAX_CHAR_PER_LINE/2

#define MAX_LENGTH_I2C_FRAME 				16

#define BOOTLOADER_CMD_PREFIX   			0xFF

#define ENTER_BOOTLOADER_CMD    			0x38
#define WRITE_BLOCK_CMD         			0x39
#define EXIT_BOOTLOADER_CMD     			0x3B
#define RUN_APP_CMD             			0xA5

#define STATE_INITIAL           			0
#define STATE_ENTER_BOOTLOADER  			1
#define STATE_WRITE_BLOCK       			2
#define STATE_EXIT_BOOTLOADER   			3

#define ERROR_INVALID_FIRMWARE_FORMAT                   10
#define ERROR_FIRMWARE_DOES_NOT_START_WITH_ENTER_BL_CMD 20
#define ERROR_INVALID_CMD_SEQUENCE                      30
#define ERROR_UNKNOWN_STATE                             40
#define ERROR_UNKNOWN_CMD                               50
#define ERROR_TS_REBOOTED				60
#define ERROR_I2C_ACCESS				70

#define SIZE_PREFIX_BOOTLOADER_CMD      		2
#define SIZE_BOOTLOADER_CMD             		2
#define SIZE_BOOTLOADER_KEY             		8

#define STATUS_REGISTER 				0

#define VERBOSE_DEBUG    				2
#define VERBOSE_ERROR    				1
#define VERBOSE_DISABLED 				0

#define DELAY_BETWEEN_RELEASE_RST_AND_SEND_1ST_CMD      100 //in ms 
#define MAX_RETRIES_BL_ACCESS				20
#define DELAY_TO_REBOOT_AFTER_UPDATE			3000 //in ms

typedef struct           
{                                                                             
	struct semaphore                mutex;                                
} psoc_drv_data_t;                                                            

typedef struct finger_position                                                
{                                                                             
	unsigned short x;                                                     
	unsigned short y;                                                     
	unsigned char z;                                                      
} __attribute__((__packed__)) FINGER_POSITION_STRUCT;                         

typedef struct i2c_data_received                                              
{                                                                             
	unsigned char hst_mode;                                               
	unsigned char tt_mode;                                                
	unsigned char large_objects_detect_and_touches;                       
	FINGER_POSITION_STRUCT finger1_position;                              
	unsigned char event_index;                                            
	FINGER_POSITION_STRUCT finger2_position;                              
	unsigned char gesture_count_and_overflow_flag;                        
        unsigned char gestureID;                                              
} __attribute__((__packed__)) I2C_DATA_RECEIVED_STRUCT;                       


typedef struct i2c_data_received_sys_info                                     
{                                                                             
	unsigned char hst_mode;                                               
	unsigned char bist_comm;                                              
	unsigned char bist_stat;                                              
	unsigned char unused1;                                                
	unsigned char unused2;                                                
	unsigned char unused3;                                                
	unsigned char unused4;                                                
	unsigned char siliconID0;                                             
	unsigned char siliconID1;                                             
	unsigned char siliconID2;                                             
	unsigned char siliconID3;                                             
	unsigned char siliconID4;                                             
	unsigned char siliconID5;                                             
	unsigned char siliconID6;                                             
	unsigned char siliconID7;                                             
	unsigned short bootloader_version;                                    
	unsigned short truetouch_standard_product_version;                    
	unsigned short applicationID;                                         
	unsigned char application_version_major;                              
	unsigned char application_version_minor;                              
	unsigned char unused5;                                                
	unsigned char unused6;                                                
	unsigned char unused7;                                                
	unsigned char unused8;                                                
	unsigned short resolution_X;                                          
	unsigned short resolution_Y;                                          
	unsigned char number_gestures;                                        
} __attribute__((__packed__)) I2C_DATA_RECEIVED_SYS_INFO_STRUCT;              


typedef struct i2c_data_received_raw_counts                                   
{                                                                             
	unsigned char hst_mode;                                               
	unsigned char tt_mode;                                                
	unsigned char tt_stat;                                                
	unsigned short first_touch_x;                                         
	unsigned short first_touch_y;                                         
	unsigned short col_data[COLUMNS];                                     
	unsigned short row_data[ROWS];                                        
} __attribute__((__packed__))  I2C_DATA_RECEIVED_RAW_COUNTS_STRUCT;

typedef struct
{
	void (*reset_ts)(void);
	char (*key);
}  psoc_pdata_t;

#define PSOC_DEVNAME			"cypress-cap-ts"
#define PSOC_301_DEVNAME		"tomtom-psoc-301"
#define PSOC_340_DEVNAME		"tomtom-psoc-340"
#define PSOC_BOOTLOADER_DEVNAME		"tomtom-psoc-bl"
#endif

